# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
import ubelt as ub
import re


class SupressPrint():
    """
    Temporarily replace the print function in a module with a noop

    Args:
        *mods: the modules to disable print in
        **kw: only accepts "enabled"
            enabled (bool, default=True): enables or disables this context
    """
    def __init__(self, *mods, **kw):
        enabled = kw.get('enabled', True)
        self.mods = mods
        self.enabled = enabled
        self.oldprints = {}

    def __enter__(self):
        if self.enabled:
            for mod in self.mods:
                oldprint = getattr(self.mods, 'print', print)
                self.oldprints[mod] = oldprint
                mod.print = lambda *args, **kw: None
    def __exit__(self, a, b, c):
        if self.enabled:
            for mod in self.mods:
                mod.print = self.oldprints[mod]


class FlatIndexer(ub.NiceRepr):
    """
    Creates a flat "view" of a jagged nested indexable object.
    Only supports one offset level.

    TODO:
        - [ ] Move to kwarray

    Args:
        lens (list): a list of the lengths of the nested objects.

    Doctest:
        >>> self = FlatIndexer([1, 2, 3])
        >>> len(self)
        >>> self.unravel(4)
        >>> self.ravel(2, 1)
    """
    def __init__(self, lens):
        self.lens = lens
        self.cums = np.cumsum(lens)

    @classmethod
    def fromlist(cls, items):
        lens = list(map(len, items))
        return cls(lens)

    def __len__(self):
        return self.cums[-1] if len(self.cums) else 0

    def unravel(self, index):
        """
        Args:
            index : raveled index

        Returns:
            Tuple[int, int]: outer and inner indices
        """
        outer = np.where(self.cums > index)[0][0]
        base = self.cums[outer] - self.lens[outer]
        inner = index - base
        return (outer, inner)

    def ravel(self, outer, inner):
        """
        Args:
            outer: index into outer list
            inner: index into the list referenced by outer

        Returns:
            index: the raveled index
        """
        base = self.cums[outer] - self.lens[outer]
        return base + inner


def strip_ansi(text):
    r"""
    Removes all ansi directives from the string.

    References:
        http://stackoverflow.com/questions/14693701/remove-ansi
        https://stackoverflow.com/questions/13506033/filtering-out-ansi-escape-sequences

    Examples:
        >>> line = '\t\u001b[0;35mBlabla\u001b[0m     \u001b[0;36m172.18.0.2\u001b[0m'
        >>> escaped_line = strip_ansi(line)
        >>> assert escaped_line == '\tBlabla     172.18.0.2'
    """
    # ansi_escape1 = re.compile(r'\x1b[^m]*m')
    # text = ansi_escape1.sub('', text)
    # ansi_escape2 = re.compile(r'\x1b\[([0-9,A-Z]{1,2}(;[0-9]{1,2})?(;[0-9]{3})?)?[m|K]?')
    ansi_escape3 = re.compile(r'(\x9B|\x1B\[)[0-?]*[ -/]*[@-~]', flags=re.IGNORECASE)
    text = ansi_escape3.sub('', text)
    return text


def align(text, character='=', replchar=None, pos=0):
    r"""
    Left justifies text on the left side of character

    Args:
        text (str): text to align
        character (str): character to align at
        replchar (str): replacement character (default=None)

    Returns:
        str: new_text

    Example:
        >>> character = '='
        >>> text = 'a = b=\none = two\nthree = fish\n'
        >>> print(text)
        >>> result = (align(text, '='))
        >>> print(result)
        a     = b=
        one   = two
        three = fish
    """
    line_list = text.splitlines()
    new_lines = align_lines(line_list, character, replchar, pos=pos)
    new_text = '\n'.join(new_lines)
    return new_text


def align_lines(line_list, character='=', replchar=None, pos=0):
    r"""
    Left justifies text on the left side of character

    align_lines

    TODO:
        clean up and move to ubelt?

    Args:
        line_list (list of strs):
        character (str):
        pos (int or list or None): does one alignment for all chars beyond this
            column position. If pos is None, then all chars are aligned.

    Returns:
        list: new_lines

    Example:
        >>> line_list = 'a = b\none = two\nthree = fish'.split('\n')
        >>> character = '='
        >>> new_lines = align_lines(line_list, character)
        >>> result = ('\n'.join(new_lines))
        >>> print(result)
        a     = b
        one   = two
        three = fish

    Example:
        >>> line_list = 'foofish:\n    a = b\n    one    = two\n    three    = fish'.split('\n')
        >>> character = '='
        >>> new_lines = align_lines(line_list, character)
        >>> result = ('\n'.join(new_lines))
        >>> print(result)
        foofish:
            a        = b
            one      = two
            three    = fish

    Example:
        >>> import ubelt as ub
        >>> character = ':'
        >>> text = ub.codeblock('''
            {'max': '1970/01/01 02:30:13',
             'mean': '1970/01/01 01:10:15',
             'min': '1970/01/01 00:01:41',
             'range': '2:28:32',
             'std': '1:13:57',}''').split('\n')
        >>> new_lines = align_lines(text, ':', ' :')
        >>> result = '\n'.join(new_lines)
        >>> print(result)
        {'max'   : '1970/01/01 02:30:13',
         'mean'  : '1970/01/01 01:10:15',
         'min'   : '1970/01/01 00:01:41',
         'range' : '2:28:32',
         'std'   : '1:13:57',}

    Example:
        >>> line_list = 'foofish:\n a = b = c\n one = two = three\nthree=4= fish'.split('\n')
        >>> character = '='
        >>> # align the second occurence of a character
        >>> new_lines = align_lines(line_list, character, pos=None)
        >>> print(('\n'.join(line_list)))
        >>> result = ('\n'.join(new_lines))
        >>> print(result)
        foofish:
         a   = b   = c
         one = two = three
        three=4    = fish
    """

    # FIXME: continue to fix ansi
    if pos is None:
        # Align all occurences
        num_pos = max([line.count(character) for line in line_list])
        pos = list(range(num_pos))

    # Allow multiple alignments
    if isinstance(pos, list):
        pos_list = pos
        # recursive calls
        new_lines = line_list
        for pos in pos_list:
            new_lines = align_lines(new_lines, character=character,
                                    replchar=replchar, pos=pos)
        return new_lines

    # base case
    if replchar is None:
        replchar = character

    # the pos-th character to align
    lpos = pos
    rpos = lpos + 1

    tup_list = [line.split(character) for line in line_list]

    handle_ansi = True
    if handle_ansi:
        # Remove ansi from length calculation
        # References: http://stackoverflow.com/questions/14693701remove-ansi
        ansi_escape = re.compile(r'\x1b[^m]*m')

    # Find how much padding is needed
    maxlen = 0
    for tup in tup_list:
        if len(tup) >= rpos + 1:
            if handle_ansi:
                tup = [ansi_escape.sub('', x) for x in tup]
            left_lenlist = list(map(len, tup[0:rpos]))
            left_len = sum(left_lenlist) + lpos * len(replchar)
            maxlen = max(maxlen, left_len)

    # Pad each line to align the pos-th occurence of the chosen character
    new_lines = []
    for tup in tup_list:
        if len(tup) >= rpos + 1:
            lhs = character.join(tup[0:rpos])
            rhs = character.join(tup[rpos:])
            # pad the new line with requested justification
            newline = lhs.ljust(maxlen) + replchar + rhs
            new_lines.append(newline)
        else:
            new_lines.append(replchar.join(tup))
    return new_lines



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import ubelt as ub
import numpy as np
import netharn as nh
import torch
import itertools as it


class SlidingWindow(ub.NiceRepr):
    """
    Slide a window of a certain shape over an array with a larger shape.

    This can be used for iterating over a grid of sub-regions of 2d-images,
    3d-volumes, or any n-dimensional array.

    Yields slices of shape `window` that can be used to index into an array
    with shape `shape` via numpy / torch fancy indexing. This allows for fast
    fast iteration over subregions of a larger image. Because we generate a
    grid-basis using only shapse, the larger image does not need to be in
    memory as long as its width/height/depth/etc...

    Args:
        shape (Tuple[int, ...]): shape of source array to slide across.

        window (Tuple[int, ...]): shape of window that will be slid over the
            larger image.

        overlap (float, default=0): a number between 0 and 1 indicating the
            fraction of overlap that parts will have. Specifying this is
            mutually exclusive with `stride`.  Must be `0 <= overlap < 1`.

        stride (int, default=None): the number of cells (pixels) moved on each
            step of the window. Mutually exclusive with overlap.

        keepbound (bool, default=False): if True, a non-uniform stride will be
            taken to ensure that the right / bottom of the image is returned as
            a slice if needed. Such a slice will not obey the overlap
            constraints.  (Defaults to False)

        allow_overshoot (bool, default=False): if False, we will raise an
            error if the window doesn't slide perfectly over the input shape.

    Attributes:
        basis_shape - shape of the grid corresponding to the number of strides
            the sliding window will take.
        basis_slices - slices that will be taken in every dimension

    Yields:
        Tuple[slice, ...]: slices used for numpy indexing, the number of slices
            in the tuple

    Notes:
        This is a simpler version of SlidingSlices

        For each dimension, we generate a basis (which defines a grid), and we
        slide over that basis.

    Example:
        >>> import netharn as nh
        >>> shape = (10, 10)
        >>> window = (5, 5)
        >>> self = nh.util.SlidingWindow(shape, window)
        >>> for i, index in enumerate(self):
        >>>     print('i={}, index={}'.format(i, index))
        i=0, index=(slice(0, 5, None), slice(0, 5, None))
        i=1, index=(slice(0, 5, None), slice(5, 10, None))
        i=2, index=(slice(5, 10, None), slice(0, 5, None))
        i=3, index=(slice(5, 10, None), slice(5, 10, None))

    Example:
        >>> shape = (16, 16)
        >>> window = (4, 4)
        >>> self = SlidingWindow(shape, window, overlap=(.5, .25))
        >>> print('self.stride = {!r}'.format(self.stride))
        self.stride = [2, 3]
        >>> list(ub.chunks(self.grid, 5))
        [[(0, 0), (0, 1), (0, 2), (0, 3), (0, 4)],
         [(1, 0), (1, 1), (1, 2), (1, 3), (1, 4)],
         [(2, 0), (2, 1), (2, 2), (2, 3), (2, 4)],
         [(3, 0), (3, 1), (3, 2), (3, 3), (3, 4)],
         [(4, 0), (4, 1), (4, 2), (4, 3), (4, 4)],
         [(5, 0), (5, 1), (5, 2), (5, 3), (5, 4)],
         [(6, 0), (6, 1), (6, 2), (6, 3), (6, 4)]]
    """
    def __init__(self, shape, window, overlap=None, stride=None,
                 keepbound=False, allow_overshoot=False):

        if overlap is None and stride is None:
            overlap = 0

        assert len(window) == len(shape), (
            'incompatible dims: {} {}'.format(len(window),
                                              len(shape)))
        # assert all(d <= D for d, D in zip(window, shape)), (
        #         'window must be smaller than shape')

        stride, overlap = self._compute_stride(overlap, stride, shape,
                                               window)

        # print('stride = {!r}'.format(stride))
        if not all(stride):
            raise ValueError(
                'Step must be positive everywhere. Got={}'.format(stride))

        stide_kw = [dict(margin=d, stop=D, step=s, keepbound=keepbound,
                         check=not keepbound and not allow_overshoot)
                      for d, D, s in zip(window, shape, stride)]

        undershot_shape = []
        overshoots = []
        for kw in stide_kw:
            final_pos = (kw['stop'] - kw['margin'])
            n_steps = final_pos // kw['step']
            overshoot = final_pos % kw['step']
            undershot_shape.append(n_steps + 1)
            overshoots.append(overshoot)

        self._final_step = overshoots

        if not allow_overshoot and any(overshoots):
            raise ValueError('overshoot={} stide_kw={}'.format(overshoots,
                                                               stide_kw))

        # make a slice generator for each dimension
        self.stride = stride
        self.overlap = overlap

        self.window = window
        self.input_shape = shape

        # The undershot basis shape, only contains indices that correspond
        # perfectly to the input. It may crop a bit of the ends.  If this is
        # equal to basis_shape, then the self perfectly fits the input.
        self.undershot_shape = undershot_shape

        # NOTE: if we have overshot, then basis shape will not perfectly
        # align to the original image. This shape will be a bit bigger.
        self.basis_slices = [list(nh.util.wide_strides_1d(**kw))
                               for kw in stide_kw]
        self.basis_shape = [len(b) for b in self.basis_slices]
        self.n_total = np.prod(self.basis_shape)

    def __nice__(self):
        return 'bshape={}, shape={}, window={}, stride={}'.format(
            tuple(self.basis_shape),
            tuple(self.input_shape),
            self.window,
            tuple(self.stride)
        )

    def _compute_stride(self, overlap, stride, shape, window):
        """
        Ensures that stride hasoverlap the correct shape.  If stride is not
        provided, compute stride from desired overlap.
        """
        if isinstance(stride, np.ndarray):
            stride = tuple(stride)
        if isinstance(overlap, np.ndarray):
            overlap = tuple(overlap)

        if not (overlap is None) ^ (stride is None):
            raise ValueError('specify overlap({}) XOR stride ({})'.format(
                overlap, stride))
        if stride is None:
            if not isinstance(overlap, (list, tuple)):
                overlap = [overlap] * len(window)
            if any(frac < 0 or frac >= 1 for frac in overlap):
                raise ValueError((
                    'part overlap was {}, but fractional overlaps must be '
                    'in the range [0, 1)').format(overlap))
            stride = [int(round(d - d * frac))
                      for frac, d in zip(overlap, window)]
        else:
            if not isinstance(stride, (list, tuple)):
                stride = [stride] * len(window)
        # Recompute fractional overlap after integer stride is computed
        overlap = [(d - s) / d for s, d in zip(stride, window)]
        assert len(stride) == len(shape), 'incompatible dims'
        return stride, overlap

    def __len__(self):
        return self.n_total

    def _iter_basis_frac(self):
        for slices in self:
            frac = [sl.start / D for sl, D in zip(slices, self.source.shape)]
            yield frac

    def __iter__(self):
        for slices in it.product(*self.basis_slices):
            yield slices

    def __getitem__(self, index):
        """
        Get a specific item by its flat (raveled) index

        Example:
            >>> import netharn as nh
            >>> shape = (220, 220)
            >>> window = (10, 10)
            >>> self = nh.util.SlidingWindow(shape, window, stride=5)
            >>> itered_items = list(self)
            >>> assert len(itered_items) == len(self)
            >>> indexed_items = [self[i] for i in range(len(self))]
            >>> assert itered_items[0] == self[0]
            >>> assert itered_items[-1] == self[-1]
            >>> assert itered_items == indexed_items
        """
        if index < 0:
            index = len(self) + index
        # Find the nd location in the grid
        basis_idx = np.unravel_index(index, self.basis_shape)
        # Take the slice for each of the n dimensions
        slices = tuple([bdim[i]
                        for bdim, i in zip(self.basis_slices, basis_idx)])
        return slices

    @property
    def grid(self):
        """
        Generate indices into the "basis" slice for each dimension.
        This enumerates the nd indices of the grid.

        Yields:
            Tuple[int, ]
        """
        # Generates basis for "sliding window" slices to break a large image
        # into smaller pieces. Use it.product to slide across the coordinates.
        basis_indices = map(range, self.basis_shape)
        for basis_idxs in it.product(*basis_indices):
            yield basis_idxs

    @property
    def slices(self):
        """
        Generate slices for each window (equivalent to iter(self))

        Example:
            >>> shape = (220, 220)
            >>> window = (10, 10)
            >>> self = SlidingWindow(shape, window, stride=5)
            >>> list(self)[41:45]
            [(slice(0, 10, None), slice(205, 215, None)),
             (slice(0, 10, None), slice(210, 220, None)),
             (slice(5, 15, None), slice(0, 10, None)),
             (slice(5, 15, None), slice(5, 15, None))]
            >>> print('self.overlap = {!r}'.format(self.overlap))
            self.overlap = [0.5, 0.5]
        """
        return iter(self)

    @property
    def centers(self):
        """
        Generate centers of each window

        Yields:
            Tuple[float, ...]: the center coordinate of the slice

        Example:
            >>> shape = (4, 4)
            >>> window = (3, 3)
            >>> self = SlidingWindow(shape, window, stride=1)
            >>> list(zip(self.centers, self.slices))
            [((1.0, 1.0), (slice(0, 3, None), slice(0, 3, None))),
             ((1.0, 2.0), (slice(0, 3, None), slice(1, 4, None))),
             ((2.0, 1.0), (slice(1, 4, None), slice(0, 3, None))),
             ((2.0, 2.0), (slice(1, 4, None), slice(1, 4, None)))]
            >>> shape = (3, 3)
            >>> window = (2, 2)
            >>> self = SlidingWindow(shape, window, stride=1)
            >>> list(zip(self.centers, self.slices))
            [((0.5, 0.5), (slice(0, 2, None), slice(0, 2, None))),
             ((0.5, 1.5), (slice(0, 2, None), slice(1, 3, None))),
             ((1.5, 0.5), (slice(1, 3, None), slice(0, 2, None))),
             ((1.5, 1.5), (slice(1, 3, None), slice(1, 3, None)))]
        """
        for slices in self:
            center = tuple(sl_.start + (sl_.stop - sl_.start - 1) / 2
                           for sl_ in slices)
            yield center


class Stitcher(ub.NiceRepr):
    """
    Restitches smaller image patches / pixels into a larger output.  This is
    used to invert the SlidingSlicer.  For semenatic segmentation the patches
    are probability chips. Overlapping chips are averaged together.

    Args:
        shape (tuple): dimensions of the large image that will be created from
            the smaller pixels or patches.

    Example:
        >>> import sys
        >>> # Build a high resolution image and slice it into chips
        >>> highres = np.random.rand(5, 200, 200).astype(np.float32)
        >>> target_shape = (1, 50, 50)
        >>> slider = SlidingSlices(highres, target_shape, overlap=(0, .5, .5))
        >>> # Show how Sticher can be used to reconstruct the original image
        >>> stitcher = Stitcher(slider.source.shape)
        >>> for sl, chip in list(slider):
        ...     stitcher.add(sl, chip)
        >>> assert stitcher.weights.max() == 4, 'some parts should be processed 4 times'
        >>> recon = stitcher.finalize()

    """
    def __init__(stitcher, shape, xpu='numpy'):
        stitcher.shape = shape
        stitcher.xpu = xpu
        if xpu == 'numpy':
            stitcher.sums = np.zeros(shape, dtype=np.float32)
            stitcher.weights = np.zeros(shape, dtype=np.float32)

            stitcher.sumview = stitcher.sums.ravel()
            stitcher.weightview = stitcher.weights.ravel()
        else:
            stitcher.sums = xpu.move(torch.zeros(shape))
            stitcher.weights = xpu.move(torch.zeros(shape))

            stitcher.sumview = stitcher.sums.view(-1)
            stitcher.weightview = stitcher.weights.view(-1)

            stitcher._cumprod = np.cumprod(list(shape[::-1][:-1]))[::-1]
            stitcher._cumprod = torch.LongTensor(np.array(stitcher._cumprod))

    def __nice__(stitcher):
        return str(stitcher.sums.shape)

    def add(stitcher, indices, patch, weight=None):
        """
        Incorporate a new (possibly overlapping) patch or pixel using a
        weighted sum.

        Args:
            indices (slice or tuple): typically a Tuple[slice] of pixels or a
                single pixel, but this can be any numpy fancy index.
            patch (ndarray): data to patch into the bigger image.
            weight (float or ndarray): weight of this patch (default to 1.0)
        """
        if weight is None:
            stitcher.sums[indices] += patch
            stitcher.weights[indices] += 1.0
        else:
            stitcher.sums[indices] += (patch * weight)
            stitcher.weights[indices] += weight

    def add_fast(stitcher, batch_idxs, values, weight=None, assume_order=True):
        """
        new faster version

        Ignore:

            stitcher = xxx_sticher
            values = xxx_np

            import ubelt
            for timer in ubelt.Timerit(10, bestof=1):
                with timer:
                    stitcher_add(stitcher, batch_idxs, values, assume_order=False)

            import ubelt
            for timer in ubelt.Timerit(10, bestof=1):
                with timer:
                    stitcher_add(stitcher, batch_idxs, values, assume_order=True)

            import ubelt
            batch_idxs_tuple = list(map(tuple, batch_idxs))
            for timer in ubelt.Timerit(10, bestof=1):
                with timer:
                    for indices, xxx in zip(batch_idxs_tuple, xxx_np):
                        xxx_sticher.add(indices, xxx)

        Example:
            >>> import sys
            >>> # Build a high resolution image and slice it into chips
            >>> frames = np.random.rand(1, 200, 100, 100).astype(np.float32)
            >>> window = (frames.shape[0], 15, 15, 15)
            >>> slider = SlidingSlices(frames, window, step=(1, 1, 1, 1))
            >>> dset = slider.to_dataset()
            >>> n_classes = 2
            >>> xpu = nh.XPU(None)
            >>> stitcher = Stitcher(slider.basis_shape[1:] + [n_classes], xpu=xpu)
            >>> loader = torch.utils.data.DataLoader(dset, batch_size=10)
            >>> batch_iter = iter(loader)
            >>> batch = next(batch_iter)
            >>> batch_idxs_tensors_, chips = batch
            >>> invar = torch.autograd.Variable(chips)
            >>> conv = torch.nn.Conv3d(frames.shape[0], n_classes, window[1:])
            >>> values = conv(invar).data
            >>> # remove channel
            >>> weight = None
            >>> batch_idxs = batch_idxs_tensors_[:, 1:]
            >>> stitcher.add_fast(batch_idxs, values, weight, assume_order=True)

        Time:
            torch.cuda.init()

            weight = None

            import ubelt as ub
            xpu = nh.XPU(0)
            values = xpu.move(values)
            stitcher = Stitcher(slider.basis_shape[1:] + [n_classes], xpu=xpu)
            for timer in ub.Timerit(100, bestof=10, label='gpu'):
                with timer:
                    stitcher.add_fast(batch_idxs, values, weight, assume_order=True)

            stitcher = Stitcher(slider.basis_shape[1:] + [n_classes], xpu='numpy')
            batch_idxs_np = batch_idxs.numpy()
            values_np = values.cpu().numpy()
            for timer in ub.Timerit(100, bestof=10, label='numpy'):
                with timer:
                    stitcher.add_fast(batch_idxs_np, values_np, weight, assume_order=True)

        Benchmark:
            >>> import sys
            >>> # setup benchmark
            >>> frames = np.random.rand(1, 50, 100, 100).astype(np.float32)
            >>> window = (frames.shape[0], 20, 20, 20)
            >>> slider = SlidingSlices(frames, window, step=(1, 1, 1, 1))
            >>> dset = slider.to_dataset()
            >>> loader = torch.utils.data.DataLoader(dset, batch_size=1024)
            >>> n_classes = 2
            >>> xpu = nh.XPU(1)
            >>> conv = torch.nn.Conv3d(window[0], n_classes, window[1:])
            >>> conv = xpu.move(conv)
            >>> #weight = torch.rand(n_classes, 1, 1, 1)[None, :]
            >>> #weight = xpu.move(weight)
            >>> #weight_np = weight.cpu().numpy()
            >>> weight = weight_np = None
            >>> # do dummy computation to warm up gpu
            >>> conv(xpu.move(dset[0][1][None, :]))
            >>> torch.set_grad_enabled(False)
            >>> conv.train(False)
            >>> base_shape = slider.basis_shape[1:]
            >>> # ---------------------------------------
            >>> # Benchmark on-gpu stitching with pytorch
            >>> import tqdm
            >>> t1 = ub.Timerit(3, bestof=3, label='gpu')
            >>> for timer in tqdm.tqdm(t1, total=3, leave=True):
            >>>     with timer:
            >>>         stitcher = Stitcher(base_shape + [n_classes], xpu=xpu)
            >>>         for batch in loader:
            >>>             batch_idxs_tensors_, chips = batch
            >>>             invar = xpu.move(chips)
            >>>             values = conv(invar).data
            >>>             batch_idxs = batch_idxs_tensors_[:, 1:].numpy()
            >>>             stitcher.add_fast(batch_idxs, values, weight,
            >>>                               assume_order=True)
            >>> # ---------------------------------------
            >>> # Benchmark on-cpu stitching with numpy
            >>> t2 = ub.Timerit(3, bestof=3, label='numpy')
            >>> for timer in tqdm.tqdm(t2, total=3, leave=True):
            >>>     with timer:
            >>>         stitcher = Stitcher(base_shape + [n_classes], xpu='numpy')
            >>>         for batch in iter(loader):
            >>>             batch_idxs_tensors_, chips = batch
            >>>             invar = xpu.move(chips)
            >>>             values_np = conv(invar).data.cpu().numpy()
            >>>             batch_idxs_np = batch_idxs_tensors_[:, 1:].numpy()
            >>>             stitcher.add_fast(batch_idxs_np, values_np,
            >>>                               weight_np, assume_order=True)
            >>> # VERDICT:
            >>> # Async GPU stitching gives a minor but insignificant speedup
            >>> # GPU:   time per loop: best=4.394 s, mean=4.394 ± 0.0 s
            >>> # NUMPY: time per loop: best=4.876 s, mean=4.876 ± 0.0 s
        """
        if stitcher.xpu != 'numpy':
            # ON GPU STITCHING
            n_classes = stitcher.shape[-1]
            end = batch_idxs.shape[0] - 1
            t_base_multi_idxs = batch_idxs[[0, end]]

            # we dont need a trailing 1 because we arent padding extra zeros
            cumprod = stitcher._cumprod[None :]
            ravel_idxs_range = (t_base_multi_idxs * cumprod).sum(dim=1)
            first = ravel_idxs_range[0]
            last = ravel_idxs_range[-1] + n_classes
            ravel_sl = slice(first, last)
            ravel_index = ravel_sl

            if weight is None:
                stitcher.sumview[ravel_index] += values.view(-1)
                stitcher.weightview[ravel_index] += 1.0
            else:
                stitcher.sumview[ravel_index] += (values * weight).view(-1)
                stitcher.weightview[ravel_index] += weight.view(-1)
        else:
            # TODO: maybe check if the input is a tensor?
            shape = stitcher.shape
            n_classes = shape[-1]
            # if we assume we get data in order, its even faster
            if assume_order:
                last = batch_idxs.shape[0] - 1
                base_multi_idxs = tuple(batch_idxs[[0, last]].T)
                # Add extra dimension for output classes
                extra_multi_idxs = np.zeros(2, dtype=np.int)
                multi_idxs_range = base_multi_idxs + (extra_multi_idxs,)
                ravel_idxs_range = np.ravel_multi_index(multi_idxs_range, dims=shape)
                first = ravel_idxs_range[0]
                last = ravel_idxs_range[-1] + n_classes
                ravel_sl = slice(first, last)
                ravel_index = ravel_sl
            else:
                base_multi_idxs = tuple(batch_idxs.T)
                extra_multi_idxs = np.zeros(len(batch_idxs), dtype=np.int)
                # The indices for the 0-th class (which should be the last dimension)
                multi_idxs_first = base_multi_idxs + (extra_multi_idxs,)
                ravel_idxs_first = np.ravel_multi_index(multi_idxs_first, dims=shape)

                # The indices for the next classes should be sequentially after
                all_ravel_idxs = [ravel_idxs_first[None, :]]
                for i in range(1, n_classes):
                    all_ravel_idxs.append((ravel_idxs_first + i)[None, :])
                # raveled indices that correspond with raveled data
                ravel_idxs = np.vstack(all_ravel_idxs).T.ravel()
                # assert np.sum(1 - np.diff(ravel_idxs)), 'we cant assume order'
                ravel_index = ravel_idxs

            if weight is None:
                stitcher.sumview[ravel_index] += values.ravel()
                stitcher.weightview[ravel_index] += 1.0
            else:
                stitcher.sumview[ravel_index] += (values * weight).ravel()
                stitcher.weightview[ravel_index] += np.ravel(weight)

    def average(stitcher):
        """
        Averages out contributions from overlapping adds using weighted average

        Returns:
            out: ndarray: the stitched image
        """
        out = stitcher.sums / stitcher.weights
        return out

    def finalize(stitcher, frame_ids=None):
        """
        Averages out contributions from overlapping adds

        Args:
            frame_ids(None or slice or tuple): if subset is not None, this is
                done for only a region of the larger tensor, otherwise it is
                done for the entire tensor.
                TODO: rename frame_ids subset

        Returns:
            final: ndarray: the stitched image
        """
        if frame_ids is None:
            final = stitcher.sums / stitcher.weights
        else:
            final = stitcher.sums[frame_ids] / stitcher.weights[frame_ids]

        if stitcher.xpu != 'numpy':
            final = final.cpu().numpy()

        final = np.nan_to_num(final)
        return final


from .util_slider_dep import SlidingIndexDataset, SlidingSlices  # NOQA


if __name__ == '__main__':
    """
    CommandLine:
        python -m netharn.util.util_slider all
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



import numpy as np
import ubelt as ub
from netharn.metrics.detections import _ave_precision
from netharn.metrics.detections import detection_confusions


def _multiclass_ap(y):
    """ computes pr like lightnet from netharn confusions """
    y = y.sort_values('score', ascending=False)

    num_annotations = y[y.true >= 0].weight.sum()
    positives = []

    for rx, row in y.iterrows():
        if row.pred >= 0:
            if row.true == row.pred:
                positives.append((row.score, True))
            else:
                positives.append((row.score, False))

    # sort matches by confidence from high to low
    positives = sorted(positives, key=lambda d: (d[0], -d[1]), reverse=True)

    tps = []
    fps = []
    tp_counter = 0
    fp_counter = 0

    # all matches in dataset
    for pos in positives:
        if pos[1]:
            tp_counter += 1
        else:
            fp_counter += 1
        tps.append(tp_counter)
        fps.append(fp_counter)

    precision = []
    recall = []
    for tp, fp in zip(tps, fps):
        recall.append(tp / num_annotations)
        precision.append(tp / np.maximum(fp + tp, np.finfo(np.float64).eps))

    import scipy
    num_of_samples = 100

    if len(precision) > 1 and len(recall) > 1:
        p = np.array(precision)
        r = np.array(recall)
        p_start = p[np.argmin(r)]
        samples = np.arange(0., 1., 1.0 / num_of_samples)
        interpolated = scipy.interpolate.interp1d(r, p, fill_value=(p_start, 0.), bounds_error=False)(samples)
        avg = sum(interpolated) / len(interpolated)
    elif len(precision) > 0 and len(recall) > 0:
        # 1 point on PR: AP is box between (0,0) and (p,r)
        avg = precision[0] * recall[0]
    else:
        avg = float('nan')

    return np.array(precision), np.array(recall), avg


def voc_eval(lines, recs, classname, ovthresh=0.5, method=False, bias=1):
    import copy
    # imagenames = ([x.strip().split(' ')[0] for x in lines])
    imagenames = ([x[0] for x in lines])
    # BUGFIX: the original code did not cast this to a set
    imagenames = set(imagenames)
    recs2 = copy.deepcopy(recs)

    # extract gt objects for this class
    class_recs = {}
    npos = 0
    for imagename in imagenames:
        R = [obj for obj in recs2[imagename] if obj['name'] == classname]
        bbox = np.array([x['bbox'] for x in R])
        difficult = np.array([x['difficult'] for x in R]).astype(np.bool)
        det = [False] * len(R)
        npos = npos + sum(~difficult)
        class_recs[imagename] = {'bbox': bbox,
                                 'difficult': difficult,
                                 'det': det}

    splitlines = lines
    image_ids = [x[0] for x in splitlines]
    confidence = np.array([x[1] for x in splitlines])
    BB = np.array([[z for z in x[2:]] for x in splitlines])

    # splitlines = [x.strip().split(' ') for x in lines]
    # confidence = np.array([float(x[1]) for x in splitlines])
    # BB = np.array([[float(z) for z in x[2:]] for x in splitlines])

    # sort by confidence
    sorted_ind = np.argsort(-confidence)
    # sorted_scores = np.sort(-confidence)  #
    BB = BB[sorted_ind, :]
    image_ids = [image_ids[x] for x in sorted_ind]

    # go down dets and mark TPs and FPs
    nd = len(image_ids)
    tp = np.zeros(nd)
    fp = np.zeros(nd)
    for d in range(nd):
        R = class_recs[image_ids[d]]
        bb = BB[d, :].astype(float)
        ovmax = -np.inf
        BBGT = R['bbox'].astype(float)

        if BBGT.size > 0:
            # compute overlaps
            # intersection
            ixmin = np.maximum(BBGT[:, 0], bb[0])
            iymin = np.maximum(BBGT[:, 1], bb[1])
            ixmax = np.minimum(BBGT[:, 2], bb[2])
            iymax = np.minimum(BBGT[:, 3], bb[3])
            iw = np.maximum(ixmax - ixmin + bias, 0.)
            ih = np.maximum(iymax - iymin + bias, 0.)
            inters = iw * ih

            # union
            uni = ((bb[2] - bb[0] + bias) * (bb[3] - bb[1] + bias) +
                   (BBGT[:, 2] - BBGT[:, 0] + bias) *
                   (BBGT[:, 3] - BBGT[:, 1] + bias) - inters)

            overlaps = inters / uni
            ovmax = np.max(overlaps)
            jmax = np.argmax(overlaps)

        if ovmax > ovthresh:
            if not R['difficult'][jmax]:
                if not R['det'][jmax]:
                    tp[d] = 1.
                    R['det'][jmax] = 1
                else:
                    fp[d] = 1.
        else:
            fp[d] = 1.

    # compute precision recall
    fp = np.cumsum(fp)
    tp = np.cumsum(tp)
    rec = tp / float(npos)
    # avoid divide by zero in case the first detection matches a difficult
    # ground truth
    prec = tp / np.maximum(tp + fp, np.finfo(np.float64).eps)

    def voc_ap(rec, prec, use_07_metric=False):
        """ ap = voc_ap(rec, prec, [use_07_metric])
        Compute VOC AP given precision and recall.
        If use_07_metric is true, uses the
        VOC 07 11 point method (default:False).
        """
        if use_07_metric:
            # 11 point metric
            ap = 0.
            for t in np.arange(0., 1.1, 0.1):
                if np.sum(rec >= t) == 0:
                    p = 0
                else:
                    p = np.max(prec[rec >= t])
                ap = ap + p / 11.
        else:
            # correct AP calculation
            # first append sentinel values at the end
            mrec = np.concatenate(([0.], rec, [1.]))
            mpre = np.concatenate(([0.], prec, [0.]))

            # compute the precision envelope
            for i in range(mpre.size - 1, 0, -1):
                mpre[i - 1] = np.maximum(mpre[i - 1], mpre[i])

            # to calculate area under PR curve, look for points
            # where X axis (recall) changes value
            i = np.where(mrec[1:] != mrec[:-1])[0]

            # and sum (\Delta recall) * prec
            ap = np.sum((mrec[i + 1] - mrec[i]) * mpre[i + 1])
        return ap

    eav_ap = voc_ap(rec, prec, use_07_metric=method ==  'voc2007')
    ap = _ave_precision(rec, prec, method)

    assert ap == eav_ap
    return rec, prec, ap


def _devcheck_voc_consistency2():
    """
    # CHECK FOR ISSUES WITH MY MAP COMPUTATION

    TODO:
        Check how cocoeval works
        https://github.com/cocodataset/cocoapi/blob/master/PythonAPI/pycocotools/cocoeval.py
    """
    import pandas as pd
    from netharn.metrics.detections import DetectionMetrics
    xdata = []
    ydatas = ub.ddict(list)

    dmets = []

    for box_noise in np.linspace(0, 8, 20):
        dmet = DetectionMetrics.demo(
            nimgs=20,
            nboxes=(0, 20),
            nclasses=3,
            rng=0,
            # anchors=np.array([[.5, .5], [.3, .3], [.1, .3], [.2, .1]]),
            box_noise=box_noise,
            # n_fp=0 if box_noise == 0 else (0, 3),
            # n_fn=0 if box_noise == 0 else (0, 3),
            # cls_noise=0 if box_noise == 0 else .3,
        )
        dmets.append(dmet)

        nh_scores = dmet.score_netharn(bias=0)
        voc_scores = dmet.score_voc(bias=0)
        coco_scores = dmet.score_coco()
        nh_map = nh_scores['mAP']
        voc_map = voc_scores['mAP']
        coco_map = coco_scores['mAP']
        print('nh_map = {!r}'.format(nh_map))
        print('voc_map = {!r}'.format(voc_map))
        print('coco_map = {!r}'.format(coco_map))

        xdata.append(box_noise)
        ydatas['voc'].append(voc_map)
        ydatas['netharn'].append(nh_map)
        ydatas['coco'].append(coco_map)

    ydf = pd.DataFrame(ydatas)
    print(ydf)

    import kwplot
    kwplot.autompl()
    kwplot.multi_plot(xdata=xdata, ydata=ydatas, fnum=1, doclf=True)

    if False:
        dmet_ = dmets[-1]
        dmet_ = dmets[0]
        print('true = ' + ub.repr2(dmet_.true.dataset, nl=2, precision=2))
        print('pred = ' + ub.repr2(dmet_.pred.dataset, nl=2, precision=2))

        dmet = DetectionMetrics()
        for gid in range(0, 5):
            print('----')
            print('gid = {!r}'.format(gid))
            dmet.true = dmet_.true.subset([gid])
            dmet.pred = dmet_.pred.subset([gid])

            nh_scores = dmet.score_netharn(bias=0)
            voc_scores = dmet.score_voc(bias=0)
            coco_scores = dmet.score_coco()
            nh_map = nh_scores['mAP']
            voc_map = voc_scores['mAP']
            coco_map = coco_scores['mAP']
            print('nh_map = {!r}'.format(nh_map))
            print('voc_map = {!r}'.format(voc_map))
            print('coco_map = {!r}'.format(coco_map))

            print('true = ' + ub.repr2(dmet.true.dataset, nl=2))
            print('pred = ' + ub.repr2(dmet.pred.dataset, nl=2))


def _devcheck_voc_consistency():
    """
    # CHECK FOR ISSUES WITH MY MAP COMPUTATION

    TODO:
        Check how cocoeval works
        https://github.com/cocodataset/cocoapi/blob/master/PythonAPI/pycocotools/cocoeval.py
    """
    import pandas as pd
    import netharn as nh
    # method = 'voc2012'
    method = 'voc2007'

    bias = 0
    bias = 0

    # classes = [0, 1, 2]
    classes = [0]

    classname = 0
    # nimgs = 5
    # nboxes = 2
    nimgs = 5
    nboxes = 5
    nbad = 1

    bg_weight=1.0
    ovthresh=0.5
    bg_cls=-1

    xdata = []
    ydatas = ub.ddict(list)
    for noise in np.linspace(0, 5, 10):
        recs = {}
        lines = []
        confusions = []
        rng = np.random.RandomState(0)

        detmetrics = DetectionMetrics()

        true_coco = nh.data.coco_api.CocoDataset()
        pred_coco = nh.data.coco_api.CocoDataset()
        cid = true_coco.add_category('cat1')
        cid = pred_coco.add_category('cat1')
        for imgname in range(nimgs):

            # Create voc style data
            imgname = str(imgname)
            import kwimage
            true_boxes = kwimage.Boxes.random(num=nboxes, scale=100., rng=rng, format='cxywh')
            pred_boxes = true_boxes.copy()
            pred_boxes.data = pred_boxes.data.astype(np.float) + (rng.rand() * noise)
            if nbad:
                pred_boxes.data = np.vstack([
                    pred_boxes.data,
                    kwimage.Boxes.random(num=nbad, scale=100., rng=rng, format='cxywh').data])

            true_cxs = rng.choice(classes, size=len(true_boxes))
            pred_cxs = true_cxs.copy()

            change = rng.rand(len(true_cxs)) < (noise / 5)
            pred_cxs_swap = rng.choice(classes, size=len(pred_cxs))
            pred_cxs[change] = pred_cxs_swap[change]
            if nbad:
                pred_cxs = np.hstack([pred_cxs, rng.choice(classes, size=nbad)])

            np.array([0] * len(true_boxes))
            pred_cxs = np.array([0] * len(pred_boxes))

            recs[imgname] = []
            for bbox in true_boxes.to_tlbr().data:
                recs[imgname].append({
                    'bbox': bbox,
                    'difficult': False,
                    'name': classname
                })

            for bbox, score in zip(pred_boxes.to_tlbr().data, np.arange(len(pred_boxes))):
                lines.append([imgname, score] + list(bbox))
                # lines.append('{} {} {} {} {} {}'.format(imgname, score, *bbox))

            # Create MS-COCO style data
            gid = true_coco.add_image(imgname)
            gid = pred_coco.add_image(imgname)
            for bbox in true_boxes.to_xywh():
                true_coco.add_annotation(gid, cid, bbox=bbox, iscrowd=False,
                                         ignore=0, area=bbox.area[0])
            for bbox, score in zip(pred_boxes.to_xywh(), np.arange(len(pred_boxes))):
                pred_coco.add_annotation(gid, cid, bbox=bbox, iscrowd=False,
                                         ignore=0, score=score,
                                         area=bbox.area[0])

            # Create netharn style confusion data
            true_weights = np.array([1] * len(true_boxes))
            pred_scores = np.arange(len(pred_boxes))

            y = pd.DataFrame(detection_confusions(true_boxes, true_cxs,
                                                  true_weights, pred_boxes,
                                                  pred_scores, pred_cxs,
                                                  bg_weight=1.0, ovthresh=0.5,
                                                  bg_cls=-1, bias=bias))
            y['gx'] = int(imgname)
            y = (y)
            confusions.append(y)

        from pycocotools import cocoeval as coco_score
        cocoGt = true_coco._aspycoco()
        cocoDt = pred_coco._aspycoco()

        evaler = coco_score.COCOeval(cocoGt, cocoDt, iouType='bbox')
        evaler.evaluate()
        evaler.accumulate()
        evaler.summarize()
        coco_ap = evaler.stats[1]

        y = pd.concat(confusions)

        mine_ap = score_detection_assignment(y, method=method)['ap']
        voc_rec, voc_prec, voc_ap = voc_eval(lines, recs, classname,
                                             ovthresh=0.5, method=method,
                                             bias=bias)
        eav_prec, eav_rec, eav_ap1 = _multiclass_ap(y)

        eav_ap2 = _ave_precision(eav_rec, eav_prec, method=method)
        voc_ap2 = _ave_precision(voc_rec, voc_prec, method=method)

        eav_ap = eav_ap2

        print('noise = {!r}'.format(noise))
        print('mine_ap = {!r}'.format(mine_ap.values.mean()))
        print('voc_ap = {!r}'.format(voc_ap))
        print('eav_ap = {!r}'.format(eav_ap))
        print('---')
        xdata.append(noise)
        ydatas['voc'].append(voc_ap)
        ydatas['eav'].append(eav_ap)
        ydatas['netharn'].append(mine_ap.values.mean())
        ydatas['coco'].append(coco_ap)

    ydf = pd.DataFrame(ydatas)
    print(ydf)

    import kwplot
    kwplot.autompl()
    kwplot.multi_plot(xdata=xdata, ydata=ydatas, fnum=1, doclf=True)



import netharn as nh
import ubelt as ub
import torch.utils
import torch
import numpy as np


class MatchingSamplerPK(ub.NiceRepr, torch.utils.data.sampler.BatchSampler):
    """
    Samples random triples from a PCC-complient dataset

    Args:
        pccs (List[FrozenSet]):
            Groups of annotation-indices, where each group contains all annots
            with the same name (individual identity).
        p (int): number of individuals sampled per batch
        k (int): number of annots sampled per individual within a batch
        batch_size (int): if specified, k is adjusted to an appropriate length
        drop_last (bool): ignored
        num_batches (int): length of the loader
        rng (int | Random, default=None): random seed
        shuffle (bool): if False rng is ignored and getitem is deterministic

    TODO:
        Look at
        https://github.com/adambielski/siamese-triplet/blob/master/datasets.py
        to see if we can try using all examples of a class before repeating
        them

    Example:
        >>> pccs = [(0, 1, 2, 3), (4, 5, 6), (7, 8, 9), (10, 11), (12,)]
        >>> batch_sampler = self = MatchingSamplerPK(pccs, p=2, k=2, shuffle=True)
        >>> print('batch_sampler = {!r}'.format(batch_sampler))
        >>> for indices in batch_sampler:
        >>>     print('indices = {!r}'.format(indices))
    """
    def __init__(self, pccs, p=21, k=4, batch_size=None, drop_last=False,
                 rng=None, shuffle=False, num_batches=None, replace=True):
        self.drop_last = drop_last
        self.replace = replace
        self.shuffle = shuffle
        self.pccs = pccs

        assert k > 1

        if replace is False:
            raise NotImplementedError(
                'We currently cant sample without replacement')

        if getattr(pccs, '__hasgraphid__', False):
            raise NotImplementedError('TODO: graphid API sampling')
        else:
            # Compute the total possible number of triplets
            # Its probably a huge number, but lets do it anyway.
            # --------
            # In the language of graphid (See Jon Crall's 2017 thesis)
            # The matching graph for MNIST is fully connected graph.  All
            # pairs of annotations with the same label have a positive edge
            # between them. All pairs of annotations with a different label
            # have a negative edge between them. There are no incomparable
            # edges. Therefore for each PCC, A the number of triples it can
            # contribute is the number of internal positive edges
            # ({len(A) choose 2}) times the number of outgoing negative edges.
            # ----
            # Each pair of positive examples could be a distinct triplet
            # For each of these any negative could be chosen
            # The number of distinct triples contributed by this PCC is the
            # product of num_pos_edges and num_neg_edges.
            import scipy
            self.num_triples = 0
            self.num_pos_edges = 0
            default_num_batches = 0
            for pcc in ub.ProgIter(self.pccs, 'pccs',  enabled=0):
                num_pos_edges = scipy.special.comb(len(pcc), 2)
                if num_pos_edges > 0:
                    default_num_batches += len(pcc)
                other_pccs = [c for c in self.pccs if c is not pcc]
                num_neg_edges = sum(len(c) for c in other_pccs)
                self.num_triples += num_pos_edges * num_neg_edges
                self.num_pos_edges += num_pos_edges

        self.multitons = [pcc for pcc in self.pccs if len(pcc) > 1]

        p = min(len(self.multitons), p)
        k = min(max(len(p) for p in self.pccs), k)
        assert k > 1

        if batch_size is not None:
            p = batch_size // k

        batch_size = p * k

        if not num_batches:
            num_batches = default_num_batches

        self.batch_size = batch_size
        self.num_batches = num_batches
        self.p = p  # PCCs per batch
        self.k = k  # Items per PCC per batch
        self.rng = nh.util.ensure_rng(rng, api='python')

    def __nice__(self):
        return ('p={p}, k={k}, batch_size={batch_size}, '
                'len={num_batches}').format(**self.__dict__)

    def __iter__(self):
        for index in range(len(self)):
            indices = self[index]
            yield indices

    def __getitem__(self, index):
        if not self.shuffle:
            self.rng = nh.util.ensure_rng(index, api='python')

        sub_pccs = self.rng.sample(self.multitons, self.p)

        groups = []
        for sub_pcc in sub_pccs:
            aids = self.rng.sample(sub_pcc, min(self.k, len(sub_pcc)))
            groups.append(aids)

        nhave = sum(map(len, groups))
        while nhave < self.batch_size:
            sub_pcc = self.rng.choice(self.pccs)
            aids = self.rng.sample(sub_pcc, min(self.k, len(sub_pcc)))
            groups.append(aids)
            nhave = sum(map(len, groups))
            overshoot = nhave - self.batch_size
            if overshoot:
                groups[-1] = groups[-1][:-overshoot]

        indices = sorted(ub.flatten(groups))
        return indices

    def __len__(self):
        return self.num_batches


class BalancedBatchSampler(
        ub.NiceRepr, torch.utils.data.sampler.BatchSampler):
    """
    A sampler for balancing classes amongst batches

    Args:
        index_to_label (List[int]): the label for each index in a dataset
        batch_size (int): number of dataset indexes for each batch
        num_batches (int | str, default='auto'): number of batches to generate
        quantile (float): interpolates between under and oversamling when
            num_batches='auto'. A value of 0 is pure undersampling, and a value
            of 1 is pure oversampling.
        shuffle (bool, default=False): if True randomize batch ordering
        drop_last (bool): unused, exists for compatibility
        rng (RandomState, default=None): random seed

    Example:
        >>> from netharn.data.batch_samplers import *  # NOQA
        >>> from netharn.data.batch_samplers import RingSampler  # NOQA
        >>> import kwarray
        >>> rng = kwarray.ensure_rng(0)
        >>> classes = ['class_{}'.format(i) for i in range(5)]
        >>> # Create a random class label for each item
        >>> index_to_label = rng.randint(0, len(classes), 100)
        >>> if 1:
        >>>     # Create a rare class
        >>>     index_to_label[0:3] = 42
        >>> quantile = 0.0
        >>> self = BalancedBatchSampler(index_to_label, batch_size=4, quantile=quantile, rng=0)
        >>> print('self.label_to_freq = {!r}'.format(self.label_to_freq))
        >>> indices = list(self)
        >>> print('indices = {!r}'.format(indices))
        >>> # Print the epoch / item label frequency per epoch
        >>> label_sequence = []
        >>> index_sequence = []
        >>> for item_indices in self:
        >>>     item_indices = np.array(item_indices)
        >>>     item_labels = index_to_label[item_indices]
        >>>     index_sequence.extend(item_indices)
        >>>     label_sequence.extend(item_labels)
        >>> label_hist = ub.dict_hist(label_sequence)
        >>> index_hist = ub.dict_hist(index_sequence)
        >>> label_hist = ub.sorted_vals(label_hist, reverse=True)
        >>> index_hist = ub.sorted_vals(index_hist, reverse=True)
        >>> index_hist = ub.dict_subset(index_hist, list(index_hist.keys())[0:5])
        >>> print('label_hist = {}'.format(ub.repr2(label_hist, nl=1)))
        >>> print('index_hist = {}'.format(ub.repr2(index_hist, nl=1)))
    """

    def __init__(self, index_to_label, batch_size=1, num_batches='auto',
                 quantile=0.5, shuffle=False, rng=None):
        import kwarray

        rng = kwarray.ensure_rng(rng, api='python')
        label_to_indices = kwarray.group_items(
            np.arange(len(index_to_label)), index_to_label)

        label_to_freq = ub.map_vals(len, label_to_indices)

        label_to_subsampler = {
            label: RingSampler(indices, shuffle=shuffle, rng=rng)
            for label, indices in label_to_indices.items()
        }

        self.label_to_freq = label_to_freq
        self.index_to_label = index_to_label
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.rng = rng
        self.label_to_indices = label_to_indices
        self.label_to_subsampler = label_to_subsampler

        if num_batches == 'auto':
            self.num_batches = self._auto_num_batches(quantile)
        else:
            self.num_batches = num_batches

        self.labels = list(self.label_to_indices.keys())

    def __nice__(self):
        return ub.repr2({
            'num_batches': self.num_batches,
            'batch_size': self.batch_size,
        }, nl=0)

    def _auto_num_batches(self, quantile):
        # Over / under sample each class depending on the balance factor
        label_freq = sorted(self.label_to_freq.values())
        # if 'idf':
        #     TODO: idf balancing
        #     N = len(self.index_to_label)
        #     label_to_idf = ub.map_vals(lambda x: N / x, self.label_to_freq)
        #     denom = sum(label_to_idf.values())
        #     label_to_prob = ub.map_vals(lambda x: x / denom, label_to_idf)
        # How many times will we sample each category?
        samples_per_label = np.quantile(label_freq, quantile)
        # Compute #items as seen per epoch, and #batches from that
        epoch_items = samples_per_label * len(label_freq)
        num_batches = max(1, int(round(epoch_items / self.batch_size)))
        return num_batches

    def __len__(self):
        return self.num_batches

    def __iter__(self):
        for index in range(self.num_batches):
            yield self[index]

    def __getitem__(self, index):
        # Choose a label for each item in the batch
        if not hasattr(self.rng, 'choices'):
            # python 3.5 support
            chosen_labels = [self.rng.choice(self.labels)
                             for _ in range(self.batch_size)]
        else:
            chosen_labels = self.rng.choices(self.labels, k=self.batch_size)
        # Count the number of items we need for each label
        label_freq = ub.dict_hist(chosen_labels)

        # Sample those indices
        batch_idxs = list(ub.flatten([
            self.label_to_subsampler[label].sample(num)
            for label, num in label_freq.items()
        ]))
        return batch_idxs


class GroupedBalancedBatchSampler(ub.NiceRepr, torch.utils.data.sampler.BatchSampler):
    """
    Show items containing less frequent categories more often

    Args:
        index_to_labels (List[Listint]]): the labels for each index in a dataset
        batch_size (int): number of dataset indexes for each batch
        num_batches (int | str, default='auto'): number of batches to generate
        shuffle (bool, default=False): if True randomize batch ordering
        drop_last (bool): unused, exists for compatibility
        rng (RandomState, default=None): random seed

    References:
        https://arxiv.org/pdf/1908.09492.pdf

    Example:
        >>> from netharn.data.batch_samplers import *  # NOQA
        >>> import kwarray
        >>> rng = kwarray.ensure_rng(0)
        >>> classes = ['class_{}'.format(i) for i in range(10)]
        >>> # Create a set of random classes for each item
        >>> index_to_labels = [rng.randint(0, len(classes), rng.randint(10))
        >>>                   for _ in range(1000)]
        >>> # Create a rare class
        >>> index_to_labels[0][0] = 42
        >>> self = GroupedBalancedBatchSampler(index_to_labels, batch_size=4)
        >>> print('self.label_to_freq = {!r}'.format(self.label_to_freq))
        >>> indices = list(self)
        >>> print('indices = {!r}'.format(indices))
        >>> # Print the epoch / item label frequency per epoch
        >>> label_sequence = []
        >>> index_sequence = []
        >>> for item_indices in self:
        >>>     item_indices = np.array(item_indices)
        >>>     item_labels = list(ub.flatten(ub.take(index_to_labels, item_indices)))
        >>>     index_sequence.extend(item_indices)
        >>>     label_sequence.extend(item_labels)
        >>> label_hist = ub.dict_hist(label_sequence)
        >>> index_hist = ub.dict_hist(index_sequence)
        >>> label_hist = ub.sorted_vals(label_hist, reverse=True)
        >>> index_hist = ub.sorted_vals(index_hist, reverse=True)
        >>> index_hist = ub.dict_subset(index_hist, list(index_hist.keys())[0:5])
        >>> print('label_hist = {}'.format(ub.repr2(label_hist, nl=1)))
        >>> print('index_hist = {}'.format(ub.repr2(index_hist, nl=1)))
    """

    def __init__(self, index_to_labels, batch_size=1, num_batches='auto',
                 shuffle=False, rng=None):
        import kwarray

        rng = kwarray.ensure_rng(rng, api='python')
        label_to_indices = ub.ddict(set)

        flat_groups = []
        for index, item_labels in enumerate(index_to_labels):
            flat_groups.extend([index] * len(item_labels))
            for label in item_labels:
                label_to_indices[label].add(index)
        flat_labels = np.hstack(index_to_labels)
        self.label_to_freq = ub.dict_hist(flat_labels)

        # Use tf-idf based scheme to compute sample probabilities
        label_to_tfidf = {}
        labels = sorted(set(flat_labels))
        for label in labels:
            index_to_tf = np.zeros(len(index_to_labels))
            for index, item_labels in enumerate(index_to_labels):
                index_to_tf[index] = (label == item_labels).sum()
            idf = len(index_to_tf) / (index_to_tf > 0).sum()
            label_to_tfidf[label] = np.maximum(index_to_tf * idf, 1)
        index_to_weight = sum(label_to_tfidf.values())
        index_to_prob = index_to_weight / index_to_weight.sum()

        self.index_to_prob = index_to_prob
        self.indices = np.arange(len(index_to_prob))

        if num_batches == 'auto':
            self.num_batches = self._auto_num_batches()
        else:
            self.num_batches = num_batches

        self.index_to_labels = index_to_labels
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.rng = kwarray.ensure_rng(rng, api='numpy')

    def __nice__(self):
        return ub.repr2({
            'num_batches': self.num_batches,
            'batch_size': self.batch_size,
            'label_to_freq': self.label_to_freq,
        }, nl=0)

    def _auto_num_batches(self):
        # The right way to calculate num samples would be using a generalized
        # solutions to the coupon collector problem, but in practice that
        # expected number of samples will be too large for imbalanced datasets.
        # Therefore we punt and simply use heuristics.
        num_batches = len(self.index_to_prob)
        # else:
        #     raise NotImplementedError(balance)
        # def nth_harmonic(n):
        #     """
        #     Example:
        #         >>> n = 10
        #         >>> want = float(sympy.harmonic(n))
        #         >>> got = nth_harmonic(n)
        #         >>> np.isclose(want, got)
        #     """
        #     return np.sum(1 / np.arange(1, n + 1))

        # def uniform_coupon_ev(n):
        #     ev = n * nth_harmonic(n)
        #     return ev

        # def uniform_coupon_ev_to_collect_k(n, k):
        #     i = np.arange(n)
        #     prob_new = (n - i + 1) / n
        #     ev_new = 1 / prob_new
        #     ev = np.sum(ev_new[0:k])
        #     return ev

        # n = 100
        # uniform_coupon_ev_to_collect_k(n, int(0.6 * n))
        # n / np.arange(1, n + 1)[::-1]
        # ev_uniform = uniform_coupon_ev(len(self.index_to_prob))
        return num_batches

    def __getitem__(self, index):
        # Hack, within each batch we are going to prevent replacement
        batch_idxs = self.rng.choice(
            self.indices, p=self.index_to_prob, replace=False,
            size=self.batch_size)
        return batch_idxs

    def __iter__(self):
        for index in range(self.num_batches):
            yield self[index]

    def __len__(self):
        return self.num_batches


class RingSampler(object):
    """
    Stateful sampling without replacement until all item are exhausted

    Example:
        >>> from netharn.data.batch_samplers import RingSampler  # NOQA
        >>> self = RingSampler(list(range(1, 4)))
        >>> sampled_items = self.sample(7)
        >>> print('sampled_items = {!r}'.format(sampled_items))
        sampled_items = array([1, 2, 3, 1, 2, 3, 1])

        >>> self = RingSampler(list(range(1, 4)), rng=0, shuffle=True)
        >>> sampled_items = self.sample(7)
        >>> print('sampled_items = {!r}'.format(sampled_items))
        sampled_items = array([3, 2, 1, 1, 3, 2, 1])
    """
    def __init__(self, items, shuffle=False, rng=None):
        import kwarray
        if len(items) == 0:
            raise Exception('no items to sample')
        self.rng = kwarray.ensure_rng(rng)
        self.items = np.array(items)
        self.shuffle = shuffle
        self.indices = np.arange(len(items))
        self._pos = None
        self.refresh()

    def refresh(self):
        import kwarray
        self._pos = 0
        if self.shuffle:
            self.indices = kwarray.shuffle(self.indices, rng=self.rng)

    def sample_indices(self, size=None):
        """
        Sample indexes into the items array
        """
        n_need = size
        if size is None:
            n_need = 1
        n_total = len(self.indices)
        idx_accum = []
        while n_need > 0:
            # Take as many as we need or as many as we have
            n_avail = (n_total - self._pos)
            n_got = min(n_need, n_avail)
            n_need -= n_got

            idxs = self.indices[self._pos:self._pos + n_got]
            idx_accum.append(idxs.copy())

            # Update state, if we have exhausted all items, then refresh
            self._pos += n_got
            if self._pos == n_total:
                self.refresh()

        sampled_idxs = np.hstack(idx_accum)
        if size is None:
            sampled_idxs = sampled_idxs[0]
        return sampled_idxs

    def sample(self, size=None):
        """
        Sample items from the items array
        """
        sampled_idxs = self.sample_indices(size)
        sampled_items = self.items[sampled_idxs]
        return sampled_items



"""
WIP

This file should contain classes (that behave like torch models), but they
implement the learning of classical learning algorithms like SVM and
RandomForest.

Deep networks are amazing at learning features. However, I don't think it's
very useful to use linear logicstic regression as a classifier. In many cases I
think an SVM or a RandomForest might produce a superior classification model,
but this has yet to be shown.

TODO:
    - [ ] Classical Abstract API
    - [ ] Integration with the FitHarn
        - [ ] How do we swap netharn's backprop+SGD with sklearn's SVM and RandomForest fit methods?
        - [ ] Netharn needs a "classical" implementation of "run".
            - [ ] Simply use the data loader to load the data
            - [ ] Defaults should encourage use with deep features.
    - [ ] RandomForest
    - [ ] SVM
"""
import ubelt as ub


class ClassicalModule(ub.NiceRepr):
    """
    Abstraction for more classical approaches to learning classifiers
    """

    def __init__(self):
        raise NotImplementedError('abstract api')

    def fit(self, dataset):
        pass


class EstimatorFactory(ub.NiceRepr):
    """
    Returns sklearn classifier

    Example:
        >>> # xdoctest: +SKIP
        >>> from netharn.models.classical import *  # NOQA
        >>> self = factory = EstimatorFactory('RF', verbose=1)
        >>> print('factory = {!r}'.format(factory))
        >>> estimator = factory()
        >>> print('estimator = {!r}'.format(estimator))
    """
    def __init__(self, clf_key='RF-OVR', **kw):
        self.clf_key = clf_key
        self.kw = kw
        self.est_kw1 = None
        self.est_kw2 = None
        self.est_kw = None
        self.est_type = None
        self.wrap_type = None
        self._normalize_attrs()

    def __nice__(self):

        kw_repr = ub.repr2(self.est_kw, nl=0)
        if len(kw_repr) > 50:
            kw_repr = ub.hash_data(kw_repr)[0:8]

        text = 'clf_key={}, est_kw={}'.format(
            self.clf_key, self.est_type, self.wrap_type, kw_repr
        )
        return text

    def __call__(self):
        clf = self._make_estimator()
        return clf

    def _normalize_attrs(self):
        tup = self.clf_key.split('-')
        wrap_type = None if len(tup) == 1 else tup[1]
        est_type = tup[0]

        self.est_type = est_type
        self.wrap_type = wrap_type

        est_kw1, est_kw2 = self._lookup_params(self.est_type)
        self.est_kw1 = est_kw1
        self.est_kw2 = est_kw2
        self.est_kw = ub.dict_union(est_kw1, est_kw2, self.kw)

    def _lookup_params(self, est_type):
        if est_type in {'RF', 'RandomForest'}:
            self.est_type = 'RF'
            est_kw1 = {
                # 'max_depth': 4,
                'bootstrap': True,
                'class_weight': None,
                'criterion': 'entropy',
                'max_features': 'sqrt',
                # 'max_features': None,
                'min_samples_leaf': 5,
                'min_samples_split': 2,
                # 'n_estimators': 64,
                'n_estimators': 256,
            }
            est_kw2 = {
                'random_state': 3915904814,
                'verbose': 0,
                'n_jobs': -1,
            }
        elif est_type in {'SVC', 'SVM'}:
            self.est_type = 'SVC'
            est_kw1 = dict(kernel='linear')
            est_kw2 = {}
        elif est_type in {'Logit', 'LogisticRegression'}:
            self.est_type = 'Logit'
            est_kw1 = {}
            est_kw2 = {}
        elif est_type in {'MLP'}:
            est_kw1 = dict(
                activation='relu', alpha=1e-05, batch_size='auto',
                beta_1=0.9, beta_2=0.999, early_stopping=False,
                epsilon=1e-08, hidden_layer_sizes=(10, 10),
                learning_rate='constant', learning_rate_init=0.001,
                max_iter=200, momentum=0.9, nesterovs_momentum=True,
                power_t=0.5, random_state=3915904814, shuffle=True,
                solver='lbfgs', tol=0.0001, validation_fraction=0.1,
                warm_start=False
            )
            est_kw2 = dict(verbose=False)
        else:
            raise KeyError('Unknown Estimator')
        return est_kw1, est_kw2

    def _make_estimator(self):
        make_estimator = self._make_est_func()
        clf = make_estimator()
        return clf

    def _make_est_func(self):
        import sklearn
        from sklearn import multiclass  # NOQA
        from sklearn import ensemble  # NOQA
        from sklearn import neural_network  # NOQA
        from sklearn import svm  # NOQA
        from sklearn import preprocessing  # NOQA
        from sklearn import pipeline  # NOQA
        from functools import partial

        wrap_type = self.wrap_type
        est_type = self.est_type

        multiclass_wrapper = {
            None: ub.identity,
            'OVR': sklearn.multiclass.OneVsRestClassifier,
            'OVO': sklearn.multiclass.OneVsOneClassifier,
        }[wrap_type]
        est_class = {
            'RF': sklearn.ensemble.RandomForestClassifier,
            'SVC': sklearn.svm.SVC,
            'Logit': partial(sklearn.linear_model.LogisticRegression, solver='lbfgs'),
            'MLP': sklearn.neural_network.MLPClassifier,
        }[est_type]

        est_kw = self.est_kw
        try:
            from sklearn.impute import SimpleImputer
            Imputer = SimpleImputer
            import numpy as np
            NAN = np.nan
        except Exception:
            from sklearn.preprocessing import Imputer
            NAN = 'NaN'
        if est_type == 'MLP':
            def make_estimator():
                pipe = sklearn.pipeline.Pipeline([
                    ('inputer', Imputer(
                        missing_values=NAN, strategy='mean')),
                    # ('scale', sklearn.preprocessing.StandardScaler),
                    ('est', est_class(**est_kw)),
                ])
                return multiclass_wrapper(pipe)
        elif est_type == 'Logit':
            def make_estimator():
                pipe = sklearn.pipeline.Pipeline([
                    ('inputer', Imputer(
                        missing_values=NAN, strategy='mean')),
                    ('est', est_class(**est_kw)),
                ])
                return multiclass_wrapper(pipe)
        else:
            def make_estimator():
                return multiclass_wrapper(est_class(**est_kw))

        return make_estimator


class ClfProblem(ClassicalModule):
    """
    Takes the place of model in Hyperparams

    CommandLine:
        xdoctest -m netharn.models.classical ClfProblem:1

    Example:
        >>> # xdoctest: +SKIP
        >>> from netharn.models.classical import *  # NOQA
        >>> self = ClfProblem.demo('SVM-OVR', verbose=10)
        >>> print('self = {!r}'.format(self))
        >>> clf = self.fit()

    Example:
        >>> # xdoctest: +SKIP
        >>> from netharn.models.classical import *  # NOQA
        >>> print(chr(10) * 3 + '------------')
        >>> clf = ClfProblem.demo('SVM-OVR', verbose=1).fit()
        >>> clf = ClfProblem.demo('SVM-OVO', verbose=1).fit()
        >>> print(chr(10) * 3 + '------------')
        >>> clf = ClfProblem.demo('RF-OVR', verbose=1).fit()
        >>> clf = ClfProblem.demo('RF-OVO', verbose=1).fit()
        >>> print(chr(10) * 3 + '------------')
        >>> clf = ClfProblem.demo('Logit-OVR', verbose=1).fit()
        >>> clf = ClfProblem.demo('Logit-OVO', verbose=1).fit()
        >>> print(chr(10) * 3 + '------------')
        >>> clf = ClfProblem.demo('MLP-OVR', verbose=1).fit()
        >>> clf = ClfProblem.demo('MLP-OVO', verbose=1).fit()
    """
    def __init__(self, clf_key='RF', data=None, **kw):
        self.factory = EstimatorFactory(clf_key, **kw)
        self.data = None

        if data is not None:
            self.connect_data(data)

    def __nice__(self):
        return '{}, data=({})'.format(
            self.factory.__nice__(),
            self.data.__nice__(),)

    def connect_data(self, data):
        self.data = data

    def fit(self):
        X = self.data.X
        y = self.data.y
        clf = self.factory()
        clf = clf.fit(X, y)
        return clf

    @classmethod
    def demo(cls, clf_key='RF', **kw):
        import sklearn.datasets
        import pandas as pd
        iris = sklearn.datasets.load_iris()

        X_df = pd.DataFrame(iris.data, columns=iris.feature_names)
        Y_df = pd.DataFrame({name: iris.target == idx
                             for idx, name in enumerate(iris.target_names)})

        classes = list(map(str, Y_df.columns))
        class_idxs = (Y_df.values * [[0, 1, 2]]).sum(axis=1)

        data = ClfData(X_df, class_idxs, classes)
        self = cls(clf_key, data=data, **kw)
        return self


class ClfData(ub.NiceRepr):
    def __init__(self, X, y, classes):
        self.X = X
        self.y = y
        self.classes = classes

    def __nice__(self):
        return '{}'.format(len(self.X))


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m netharn.models.classical all
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



"""
mkinit ~/code/netharn/netharn/export
"""
from netharn.export import deployer
from netharn.export import exporter

from netharn.export.deployer import (DeployedModel,)
from netharn.export.exporter import (export_model_code,)

__all__ = ['DeployedModel', 'deployer', 'export_model_code', 'exporter']



# -*- coding: utf-8 -*-
"""
Export component of the Pytorch exporter.

This is the code that simply exports the model toplogy via code

Uses static analysis to export relevant code that defines the model topology
into a stanadlone file. As long as your model definition is indepenent of your
training code, then the exported file can be passed around in a similar way to
a caffe prototext file.

TODO:
    - [ ]: Look into: https://www.reddit.com/r/MachineLearning/comments/a856oe/d_pytorch_10_deployment_pipeline/ec9w94c/

    >>> from torchvision.models import densenet
    >>> import torch
    >>> model = densenet.DenseNet(growth_rate=16).eval()
    >>> traced = torch.jit.trace(model, example_inputs=(torch.randn(2, 3, 224, 224), ))
    >>> traced.save("densenet.pt")
    >>> model_ = torch.jit.load("densenet.pt")


CommandLine:
    xdoctest -m netharn.export.exporter export_model_code
    xdoctest -m netharn.export.exporter source_closure:1

    xdoctest -m netharn.export.exporter all
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import ast
import six  # NOQA
import re
import hashlib
import io
import pickle
import tokenize
import ubelt as ub
import warnings
from os.path import join
from . import closer

__all__ = ['export_model_code']


__pt_export_version__ = '0.5.1'


def export_model_code(dpath, model, initkw=None, export_modules=[]):
    """
    Exports the class used to define a pytorch model as a new python module.

    Exports the minimum amount of code needed to make a self-contained Python
    module defining the pytorch model class. This exports the actual source
    code. The advantage of using this over pickle is that the original code can
    change arbitrarilly because all dependencies on the original code are
    removed in the exported code.

    Args:
        dpath (str): directory to dump the model
        model (tuple or type or object): class or class instance (e.g. torch.nn.Module)
        name (str): name to use for the file (defaults to the classname)
        initkw (dict): if specified, creates the function `make`, which
            initializes the network with the specific arguments.
        export_modules (List[str]): A list of modules that the exported code
            should not depend on. Any code referenced from these modules will
            be statically extracted and copied into the model definition.
            Note that this feature is experimental.

    Returns:
        str: static_modpath: path to the saved model file.
            While you could put the output path in your PYTHONPATH, it is best
            to use `ub.import_module_from_path` to "load" the model instead.

    Example:
        >>> from netharn.export.exporter import export_model_code
        >>> from torchvision.models import densenet
        >>> import torchvision
        >>> from os.path import basename
        >>> initkw = {'growth_rate': 16}
        >>> model = densenet.DenseNet(**initkw)
        >>> dpath = ub.ensure_app_cache_dir('netharn/tests')
        >>> static_modpath = export_model_code(dpath, model, initkw)
        >>> print('static_modpath = {!r}'.format(static_modpath))
        ...
        >>> mod_fname = (basename(static_modpath))
        >>> print('mod_fname = {!r}'.format(mod_fname))
        >>> if torchvision.__version__ == '0.2.2':
        >>>     if six.PY2:
        >>>         assert mod_fname == 'DenseNet_b7ec43.py', 'got={}'.format(mod_fname)
        >>>     else:
        >>>         assert mod_fname == 'DenseNet_256629.py', 'got={}'.format(mod_fname)
        >>> # now the module can be loaded
        >>> module = ub.import_module_from_path(static_modpath)
        >>> loaded = module.make()
        >>> assert model.features.denseblock1.denselayer1.conv2.out_channels == 16
        >>> assert loaded.features.denseblock1.denselayer1.conv2.out_channels == 16
        >>> assert model is not loaded
    """
    if isinstance(model, type):
        model_class = model
    else:
        model_class = model.__class__
    classname = model_class.__name__

    if initkw is None:
        raise NotImplementedError(
            'ERROR: The params passed to the model __init__ must be available')
        footer = ''
    else:
        # First see if we can get away with a simple encoding of initkw
        try:
            # Do not use repr. The text produced is non-deterministic for
            # dictionaries. Instead, use ub.repr2, which is deterministic.
            init_text = ub.repr2(initkw, nl=1)
            eval(init_text, {})
            init_code = ub.codeblock(
                'initkw = {}'
            ).format(init_text)
        except Exception:
            # fallback to pickle
            warnings.warn('Initialization params might not be serialized '
                          'deterministically')
            init_bytes = repr(pickle.dumps(initkw, protocol=0))
            init_code = ub.codeblock(
                '''
                import pickle
                initkw = pickle.loads({})
                '''
            ).format(init_bytes)
        init_code = ub.indent(init_code).lstrip()
        # create a function to instanciate the class
        footer = '\n\n' + ub.codeblock(
            '''
            __pt_export_version__ = '{__pt_export_version__}'


            def get_initkw():
                """ creates an instance of the model """
                {init_code}
                return initkw


            def get_model_cls():
                model_cls = {classname}
                return model_cls


            def make():
                """ creates an instance of the model """
                initkw = get_initkw()
                model_cls = get_model_cls()
                model = model_cls(**initkw)
                return model
            '''
        ).format(classname=classname, init_code=init_code,
                 __pt_export_version__=__pt_export_version__)

        # TODO: assert that the name "make" is not used in the model body

    body = closer.source_closure(model_class, expand_names=export_modules)

    body_footer = body + footer + '\n'
    # dont need to hash the header, because comments are removed anyway

    # with open('debug-closer.py', 'w') as file:
    #     file.write(body_footer)
    hashid = hash_code(body_footer)

    header = ub.codeblock(
        '''
        """
        This module was autogenerated by netharn/export/exporter.py
        original_module={}
        classname={}
        timestamp={}
        hashid={}
        """
        ''').format(model_class.__module__, classname, ub.timestamp(), hashid)

    sourcecode = header + '\n' + body_footer

    static_modname = classname + '_' + hashid[0:6]
    static_modpath = join(dpath, static_modname + '.py')
    with open(static_modpath, 'w') as file:
        file.write(sourcecode)
    return static_modpath


def remove_comments_and_docstrings(source):
    r"""
    Args:
        source (str): uft8 text of source code

    Returns:
        str: out: the source with comments and docstrings removed.

    References:
        https://stackoverflow.com/questions/1769332/remove-comments-docstrings

    Example:
        >>> source = ub.codeblock(
            '''
            def foo():
                'The spaces before this docstring are tokenize.INDENT'
                test = [
                    'The spaces before this string do not get a token'
                ]
            ''')
        >>> out = remove_comments_and_docstrings(source)
        >>> want = ub.codeblock(
            '''
            def foo():
                pass
                test = [
                    'The spaces before this string do not get a token'
                ]''').splitlines()
        >>> got = [o.rstrip() for o in out.splitlines()]
        >>> assert got == want


        >>> source = ub.codeblock(
            '''
            def foo():
                " docstring "
            ''')
        >>> out = remove_comments_and_docstrings(source)
        >>> print(out)
        >>> source = ub.codeblock(
            '''
            class foo():
                r{qqq}
                docstring
                {qqq}
            ''').format(qqq='"' * 3)
        >>> out = remove_comments_and_docstrings(source)
        >>> print(out)

    """
    source = ub.ensure_unicode(source)
    io_obj = io.StringIO(source)
    output_parts = []
    prev_toktype = tokenize.INDENT
    last_lineno = -1
    last_col = 0
    for tok in tokenize.generate_tokens(io_obj.readline):
        token_type = tok[0]
        token_string = tok[1]
        start_line, start_col = tok[2]
        end_line, end_col = tok[3]
        # ltext = tok[4]
        # The following two conditionals preserve indentation.
        # This is necessary because we're not using tokenize.untokenize()
        # (because it spits out code with copious amounts of oddly-placed
        # whitespace).
        if start_line > last_lineno:
            last_col = 0
        if start_col > last_col:
            output_parts.append((' ' * (start_col - last_col)))
        # Remove comments:
        if token_type == tokenize.COMMENT:
            pass
        # This series of conditionals removes docstrings:
        elif token_type == tokenize.STRING:
            if prev_toktype != tokenize.INDENT:
                # This is likely a docstring; double-check we're not inside an
                # operator:
                if prev_toktype != tokenize.NEWLINE:
                    # Note regarding NEWLINE vs NL: The tokenize module
                    # differentiates between newlines that start a new statement
                    # and newlines inside of operators such as parens, brackes,
                    # and curly braces.  Newlines inside of operators are
                    # NEWLINE and newlines that start new code are NL.
                    # Catch whole-module docstrings:
                    if start_col > 0:
                        # Unlabelled indentation means we're inside an operator
                        output_parts.append(token_string)
                    # Note regarding the INDENT token: The tokenize module does
                    # not label indentation inside of an operator (parens,
                    # brackets, and curly braces) as actual indentation.
            else:
                # NOTE: simply removing docstrings may create invalid code
                # in cases where the only body is a docstring (e.g. a
                # custom exception). Insert a pass to prevent this.  It
                # would be nice to detect when this is necessary.
                output_parts.append('pass')
        else:
            output_parts.append(token_string)
        prev_toktype = token_type
        last_col = end_col
        last_lineno = end_line
    out = ''.join(output_parts)
    return out


def hash_code(sourcecode):
    r"""
    Hashes source code text, but tries to normalize things like whitespace and
    comments, so very minor changes wont change the hash.

    Args:
        source (str): uft8 text of source code

    Returns:
        str: hashid: 128 character (512 byte) hash of the normalized input

    Notes:
        The return value of this function is based on the AST parse tree, which
        might change between different version of Python. However, within the
        same version of Python, the results should be consistent.

    CommandLine:
        xdoctest -m /home/joncrall/code/netharn/netharn/export/exporter.py hash_code

    Example:
        >>> hashid1 = (hash_code('x = 1')[0:8])
        >>> hashid2 = (hash_code('x=1 # comments and spaces dont matter')[0:8])
        >>> hashid3 = (hash_code('\nx=1')[0:8])
        >>> assert ub.allsame([hashid1, hashid2, hashid3])
        >>> hashid4 = hash_code('x=2')[0:8]
        >>> assert hashid1 != hashid4
    """
    # Strip docstrings before making a parse tree
    sourcecode = ub.ensure_unicode(sourcecode)

    stripped = remove_comments_and_docstrings(sourcecode)

    # Also remove pytorch_export version info (not sure if correct?)
    stripped = re.sub('__pt_export_version__ = .*', '', stripped)

    parse_tree = ast.parse(stripped)
    # hashing the parse tree will normalize for a lot possible small changes
    ast_dump = ast.dump(parse_tree)

    hasher = hashlib.sha512()
    hasher.update(ast_dump.encode('utf8'))
    hashid = hasher.hexdigest()
    return hashid


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m netharn.export.exporter
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



# -*- coding: utf-8 -*-
"""
Implementation taken from [1] and [2]

References:
    [1] https://raw.githubusercontent.com/warmspringwinds/pytorch-segmentation-detection/master/pytorch_segmentation_detection/models/psp.py
    [2] https://raw.githubusercontent.com/warmspringwinds/vision/eb6c13d3972662c55e752ce7a376ab26a1546fb5/torchvision/models/resnet.py
    https://paperswithcode.com/sota/semantic-segmentation-on-camvid
    https://github.com/warmspringwinds/pytorch-segmentation-detection/tree/master/pytorch_segmentation_detection/models
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import torch.nn as nn
import torch
import math
import torch.utils.model_zoo as model_zoo
import numpy as np
from netharn import layers


# __all__ = ['PSPNet_Resnet50_8s']


model_urls = {
    'resnet18': 'https://download.pytorch.org/models/resnet18-5c106cde.pth',
    'resnet34': 'https://download.pytorch.org/models/resnet34-333f7ec4.pth',
    'resnet50': 'https://download.pytorch.org/models/resnet50-19c8e357.pth',
    'resnet101': 'https://download.pytorch.org/models/resnet101-5d3b4d8f.pth',
    'resnet152': 'https://download.pytorch.org/models/resnet152-b121ed2d.pth',
}


def conv3x3(in_planes, out_planes, stride=1, dilation=1):
    "3x3 convolution with padding"
    kernel_size = np.asarray((3, 3))
    # Compute the size of the upsampled filter with
    # a specified dilation rate.
    upsampled_kernel_size = (kernel_size - 1) * (dilation - 1) + kernel_size
    # Determine the padding that is necessary for full padding,
    # meaning the output spatial size is equal to input spatial size
    full_padding = (upsampled_kernel_size - 1) // 2
    # Conv2d doesn't accept numpy arrays as arguments
    full_padding, kernel_size = tuple(full_padding), tuple(kernel_size)
    return nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride,
                     padding=full_padding, dilation=dilation, bias=False)


class ResidualJoin(layers.Module):
    """
    Joins the main channel with a residual skip connection
    """

    def __init__(self, out_planes, inplace=True):
        super(ResidualJoin, self).__init__()
        self.inplace = inplace
        self.out_planes = out_planes

    def forward(self, main, skip):
        if self.inplace:
            out = main
            out += skip
        else:
            out = main + skip
        return out

    def output_shape_for(self, main_shape, skip_shape):
        """
        note: two inputs is not a standard yet
        """
        bsize, dims = main_shape[0], main_shape[2:]
        skip_bsize, skip_dims = skip_shape[0], skip_shape[2:]
        if bsize != skip_bsize:
            raise AssertionError('inconsistent bsize {} -vs- {}'.format(dims, skip_bsize))
        if dims != skip_dims:
            raise AssertionError('inconsistent dims {} -vs- {}'.format(dims, skip_dims))
        output_shape = tuple([bsize, self.out_planes] + list(dims))
        return output_shape

    def receptive_field_for(self, main_field, skip_field):
        """
        note: two inputs is not a standard yet
        """
        import netharn as nh
        # Handle skip connection
        # Unsure if it is ok for these values to not agree
        assert np.all(skip_field['crop'] == main_field['crop']), 'main_field={}, skip_field={}'.format(main_field, skip_field)
        assert np.all(skip_field['stride'] == main_field['stride']), 'main_field={}, skip_field={}'.format(main_field, skip_field)
        # It is definately true that maximum should be used to rectify
        # theoretical RF sizes, because RF-window of the smaller one (if they
        # are different) will be a subset of the larger one.
        field = nh.ReceptiveField({
            'shape': np.maximum(skip_field['shape'], main_field['shape']),
            'crop': np.maximum(skip_field['crop'], main_field['crop']),
            'stride': np.maximum(skip_field['stride'], main_field['stride']),
        })
        return field


class BasicBlock(layers.AnalyticModule):
    """
    Example:
        >>> from netharn.models.psp import *  # NOQA
        >>> self = BasicBlock(8, 8)
        >>> inputs = torch.rand(2, 8, 7, 7)
        >>> out = self(inputs)
    """
    expansion = 1

    def __init__(self, inplanes, planes, stride=1,
                 downsample=None, dilation=1):
        super(BasicBlock, self).__init__()
        if downsample is None:
            downsample = layers.Identity()
            pass

        self.downsample = downsample
        self.conv1 = conv3x3(inplanes, planes, stride, dilation=dilation)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes, dilation=dilation)
        self.bn2 = nn.BatchNorm2d(planes)
        self.join = ResidualJoin(planes, inplace=True)
        self.stride = stride

    def forward(self, inputs):
        out = self.conv1(inputs)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is None:
            residual = inputs
        else:
            residual = self.downsample(inputs)

        out = self.join(out, residual)
        out = self.relu(out)
        return out

    def _analytic_forward(self, inputs, _OutputFor, _Output, _Hidden,
                          **kwargs):
        """
        analytic computation of forward / shape / receptive field.
        """
        hidden = _Hidden()
        out = inputs
        out = hidden['conv1'] = _OutputFor(self.conv1)(out)
        out = hidden['bn1'] = _OutputFor(self.bn1)(out)
        out = hidden['relu1'] = _OutputFor(self.relu)(out)

        out = hidden['conv2'] = _OutputFor(self.conv2)(out)
        out = hidden['bn2'] = _OutputFor(self.bn2)(out)

        if self.downsample is None:
            residual = inputs
        else:
            residual = hidden['residual'] = _OutputFor(self.downsample)(inputs)

        out = hidden['join'] = _OutputFor(self.join)(out, residual)
        out = hidden['relu3'] = _OutputFor(self.relu)(out)
        return _Output.coerce(out, hidden)


class Bottleneck(layers.AnalyticModule):
    """
    Example:
        >>> from netharn.models.psp import *  # NOQA
        >>> self = Bottleneck(8, 2)
        >>> inputs = torch.rand(2, 8, 7, 7)
        >>> out = self(inputs)
        >>> self.output_shape_for((2, 8, 7, 7))
        >>> import netharn as nh
        >>> self.receptive_field_for()
    """
    expansion = 4

    def __init__(self, inplanes, planes, stride=1,
                 downsample=None, dilation=1):
        super(Bottleneck, self).__init__()
        if downsample is None:
            # downsample = layers.Identity()
            downsample = None

        self.downsample = downsample
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)

        self.conv2 = conv3x3(planes, planes, stride=stride, dilation=dilation)

        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * 4)
        self.relu = nn.ReLU(inplace=True)

        self.join = ResidualJoin(planes * 4, inplace=True)
        self.stride = stride

    def forward(self, inputs):
        out = self.conv1(inputs)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is None:
            residual = inputs
        else:
            residual = self.downsample(inputs)
        out = self.join(out, residual)

        out = self.relu(out)
        return out

    def _analytic_forward(self, inputs, _OutputFor, _Output, _Hidden,
                          **kwargs):
        """
        analytic computation of forward / shape / receptive field.
        """
        hidden = _Hidden()
        out = inputs
        out = hidden['conv1'] = _OutputFor(self.conv1)(out)
        out = hidden['bn1'] = _OutputFor(self.bn1)(out)
        out = hidden['relu1'] = _OutputFor(self.relu)(out)

        out = hidden['conv2'] = _OutputFor(self.conv2)(out)
        out = hidden['bn2'] = _OutputFor(self.bn2)(out)
        out = hidden['relu2'] = _OutputFor(self.relu)(out)

        out = hidden['conv3'] = _OutputFor(self.conv3)(out)
        out = hidden['bn3'] = _OutputFor(self.bn3)(out)

        if self.downsample is None:
            residual = inputs
        else:
            residual = hidden['residual'] = _OutputFor(self.downsample)(inputs)

        out = hidden['join'] = _OutputFor(self.join)(out, residual)
        out = hidden['relu3'] = _OutputFor(self.relu)(out)
        return _Output.coerce(out, hidden)


class ResNet(nn.Module):

    def __init__(self, block, layers, num_classes=1000, fully_conv=False,
                 remove_avg_pool_layer=False, output_stride=32,
                 additional_blocks=0, multi_grid=(1, 1, 1)):

        # Add additional variables to track
        # output stride. Necessary to achieve
        # specified output stride.
        self.output_stride = output_stride
        self.current_stride = 4
        self.current_dilation = 1

        self.remove_avg_pool_layer = remove_avg_pool_layer

        self.inplanes = 64
        self.fully_conv = fully_conv
        super(ResNet, self).__init__()
        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2)
        self.layer4 = self._make_layer(
            block, 512, layers[3], stride=2, multi_grid=multi_grid)

        self.additional_blocks = additional_blocks

        if additional_blocks == 1:

            self.layer5 = self._make_layer(
                block, 512, layers[3], stride=2, multi_grid=multi_grid)

        if additional_blocks == 2:

            self.layer5 = self._make_layer(
                block, 512, layers[3], stride=2, multi_grid=multi_grid)
            self.layer6 = self._make_layer(
                block, 512, layers[3], stride=2, multi_grid=multi_grid)

        if additional_blocks == 3:

            self.layer5 = self._make_layer(
                block, 512, layers[3], stride=2, multi_grid=multi_grid)
            self.layer6 = self._make_layer(
                block, 512, layers[3], stride=2, multi_grid=multi_grid)
            self.layer7 = self._make_layer(
                block, 512, layers[3], stride=2, multi_grid=multi_grid)

        self.avgpool = nn.AvgPool2d(7)

        self.fc = nn.Linear(512 * block.expansion, num_classes)

        if self.fully_conv:
            self.avgpool = nn.AvgPool2d(7, padding=3, stride=1)
            # In the latest unstable torch 4.0 the tensor.copy_
            # method was changed and doesn't work as it used to be
            #self.fc = nn.Conv2d(512 * block.expansion, num_classes, 1)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def _make_layer(self, block, planes, blocks, stride=1, multi_grid=None):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            # Check if we already achieved desired output stride.
            if self.current_stride == self.output_stride:
                # If so, replace subsampling with a dilation to preserve
                # current spatial resolution.
                self.current_dilation = self.current_dilation * stride
                stride = 1
            else:
                # If not, perform subsampling and update current
                # new output stride.
                self.current_stride = self.current_stride * stride
            # We don't dilate 1x1 convolution.
            downsample = layers.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion),
            )

        parts = []

        dilation = (multi_grid[0] * self.current_dilation
                    if multi_grid else self.current_dilation)
        parts.append(
            block(self.inplanes, planes, stride, downsample,
                  dilation=dilation))

        self.inplanes = planes * block.expansion

        for i in range(1, blocks):
            dilation = multi_grid[i] * \
                self.current_dilation if multi_grid else self.current_dilation
            parts.append(block(self.inplanes, planes, dilation=dilation))

        return layers.Sequential(*parts)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        if self.additional_blocks == 1:
            x = self.layer5(x)

        if self.additional_blocks == 2:
            x = self.layer5(x)
            x = self.layer6(x)

        if self.additional_blocks == 3:
            x = self.layer5(x)
            x = self.layer6(x)
            x = self.layer7(x)

        if not self.remove_avg_pool_layer:
            x = self.avgpool(x)

        if not self.fully_conv:
            x = x.view(x.size(0), -1)

        x = self.fc(x)
        return x

    def _analytic_forward(self, inputs, _OutputFor, _Output, _Hidden,
                          **kwargs):
        """
        analytic computation of forward / shape / receptive field.
        """
        hidden = _Hidden()
        out = inputs
        # out = hidden['conv1'] = _OutputFor(self.conv1)(out)
        # out = hidden['bn1'] = _OutputFor(self.bn1)(out)
        # out = hidden['relu1'] = _OutputFor(self.relu)(out)

        # out = hidden['conv2'] = _OutputFor(self.conv2)(out)
        # out = hidden['bn2'] = _OutputFor(self.bn2)(out)
        # out = hidden['relu2'] = _OutputFor(self.relu)(out)

        # out = hidden['conv3'] = _OutputFor(self.conv3)(out)
        # out = hidden['bn3'] = _OutputFor(self.bn3)(out)

        # residual = hidden['residual'] = _OutputFor(self.downsample)(inputs)

        # out = hidden['join'] = _OutputFor(self.join)(out, residual)
        # out = hidden['relu3'] = _OutputFor(self.relu)(out)
        return _Output.coerce(out, hidden)


def resnet18(pretrained=False, **kwargs):
    """Constructs a ResNet-18 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(BasicBlock, [2, 2, 2, 2], **kwargs)
    if pretrained:
        if model.additional_blocks:
            model.load_state_dict(
                model_zoo.load_url(
                    model_urls['resnet18']),
                strict=False)
            return model
        model.load_state_dict(model_zoo.load_url(model_urls['resnet18']))
    return model


def resnet34(pretrained=False, **kwargs):
    """Constructs a ResNet-34 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(BasicBlock, [3, 4, 6, 3], **kwargs)
    if pretrained:
        if model.additional_blocks:
            model.load_state_dict(
                model_zoo.load_url(
                    model_urls['resnet34']),
                strict=False)
            return model
        model.load_state_dict(model_zoo.load_url(model_urls['resnet34']))
    return model


def resnet50(pretrained=False, **kwargs):
    """Constructs a ResNet-50 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(Bottleneck, [3, 4, 6, 3], **kwargs)
    if pretrained:
        if model.additional_blocks:
            model.load_state_dict(
                model_zoo.load_url(model_urls['resnet50']), strict=False)
            return model
        model.load_state_dict(model_zoo.load_url(model_urls['resnet50']))
    return model


def resnet101(pretrained=False, **kwargs):
    """Constructs a ResNet-101 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(Bottleneck, [3, 4, 23, 3], **kwargs)
    if pretrained:
        if model.additional_blocks:
            model.load_state_dict(
                model_zoo.load_url(
                    model_urls['resnet101']),
                strict=False)
            return model
        model.load_state_dict(model_zoo.load_url(model_urls['resnet101']))
    return model


def resnet152(pretrained=False, **kwargs):
    """Constructs a ResNet-152 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(Bottleneck, [3, 8, 36, 3], **kwargs)
    if pretrained:
        model.load_state_dict(model_zoo.load_url(model_urls['resnet152']))
    return model


# Above is their custom resnet code


class PSP_head(layers.AnalyticModule):
    """
    Example:
        >>> from netharn.models.psp import *  # NOQA
        >>> self = PSP_head(8).eval()
        >>> inputs = torch.rand(1, 8, 10, 10)
        >>> outputs = self(inputs)
    """

    def __init__(self, in_channels, dropout=0.1):

        super(PSP_head, self).__init__()

        out_channels = int(in_channels / 4)

        Sequential = layers.Sequential

        def conv1x1_norm_noli():
            return Sequential(
                nn.Conv2d(in_channels, out_channels, 1, bias=False),
                nn.BatchNorm2d(out_channels),
                nn.ReLU(True))

        self.conv1 = conv1x1_norm_noli()
        self.conv2 = conv1x1_norm_noli()
        self.conv3 = conv1x1_norm_noli()
        self.conv4 = conv1x1_norm_noli()

        self.fusion_bottleneck = Sequential(
            nn.Conv2d(in_channels * 2, out_channels, 3, padding=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(True),
            nn.Dropout2d(dropout, False)
        )

        self.adaptive_pool1 = nn.AdaptiveAvgPool2d((1, 1))
        self.adaptive_pool2 = nn.AdaptiveAvgPool2d((2, 2))
        self.adaptive_pool3 = nn.AdaptiveAvgPool2d((3, 3))
        self.adaptive_pool4 = nn.AdaptiveAvgPool2d((6, 6))

        self._reset_weights()

    def _reset_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def forward(self, inputs):
        x = inputs
        fcn_features_spatial_dim = x.size()[2:]

        interpkw = dict(size=fcn_features_spatial_dim, mode='bilinear',
                        align_corners=True)
        interpolate = nn.functional.interpolate

        pooled_1 = self.adaptive_pool1(x)
        pooled_1 = self.conv1(pooled_1)
        pooled_1 = interpolate(pooled_1, **interpkw)

        pooled_2 = self.adaptive_pool2(x)
        pooled_2 = self.conv2(pooled_2)
        pooled_2 = interpolate(pooled_2, **interpkw)

        pooled_3 = self.adaptive_pool3(x)
        pooled_3 = self.conv3(pooled_3)
        pooled_3 = interpolate(pooled_3, **interpkw)

        pooled_4 = self.adaptive_pool4(x)
        pooled_4 = self.conv4(pooled_4)
        pooled_4 = interpolate(pooled_4, **interpkw)

        x = torch.cat([x, pooled_1, pooled_2, pooled_3, pooled_4], dim=1)
        x = self.fusion_bottleneck(x)
        return x

    def _analytic_forward(self, inputs, _OutputFor, _Output, _Hidden,
                          **kwargs):
        """
        analytic computation of forward / shape / receptive field.

        Ignore:
            >>> from netharn.models.psp import *  # NOQA
            >>> import netharn as nh
            >>> self = PSP_head(8).eval()
            >>> inputs = (1, 128, 64, 64)
            >>> _OutputFor = nh.OutputShapeFor
            >>> _Output = nh.OutputShape
            >>> _Hidden = nh.HiddenShapes
            >>> out = self.output_shape_for(inputs)
            >>> hidden = out.hidden
            >>> print(ub.repr2(hidden.shallow(1), nl=-1))

            >>> # out = self.receptive_field_for(inputs)
        """
        hidden = _Hidden()
        x = inputs

        # hack for output shape with upsample
        fcn_features_spatial_dim = inputs[2:]
        interpkw = dict(size=fcn_features_spatial_dim, mode='bilinear',
                        align_corners=True)
        interpolate = nn.functional.interpolate

        pooled_1 = x
        pooled_1 = hidden['adaptive_pool1'] = _OutputFor(self.adaptive_pool1)(pooled_1)
        pooled_1 = hidden['conv1'] = _OutputFor(self.conv1)(pooled_1)
        pooled_1 = hidden['upsample1'] = _OutputFor(interpolate)(pooled_1, **interpkw)

        pooled_2 = x
        pooled_2 = hidden['adaptive_pool2'] = _OutputFor(self.adaptive_pool2)(pooled_2)
        pooled_2 = hidden['conv2'] = _OutputFor(self.conv2)(pooled_2)
        pooled_2 = hidden['upsample2'] = _OutputFor(interpolate)(pooled_2, **interpkw)

        pooled_3 = x
        pooled_3 = hidden['adaptive_pool3'] = _OutputFor(self.adaptive_pool3)(pooled_3)
        pooled_3 = hidden['conv3'] = _OutputFor(self.conv3)(pooled_3)
        pooled_3 = hidden['upsample3'] = _OutputFor(interpolate)(pooled_3, **interpkw)

        pooled_4 = x
        pooled_4 = hidden['adaptive_pool4'] = _OutputFor(self.adaptive_pool4)(pooled_4)
        pooled_4 = hidden['conv4'] = _OutputFor(self.conv4)(pooled_4)
        pooled_4 = hidden['upsample4'] = _OutputFor(interpolate)(pooled_4, **interpkw)

        out = hidden['cat'] = _OutputFor(torch.cat)([x, pooled_1, pooled_2, pooled_3, pooled_4], dim=1)
        return _Output.coerce(out, hidden)


class PSPNet_Resnet50_8s(layers.AnalyticModule):
    """
    References:
        [1] https://arxiv.org/pdf/1612.01105.pdf

    Notes:
        # Excerpts from Section 3 in.

        The pyramid pooling module fuses features under four different pyramid
        scales.

        The coarsest level highlighted in red is global pooling to generate a single bin output.
        The following pyramid level separates the feature map into different sub-regions and forms pooled representation for different locations.

        The output of different levels in the pyramid pooling module contains the feature map with varied sizes.
        To maintain the weight of global feature, we use 1×1 convolution layer after each pyramid level to reduce the dimension of context representation to 1/N of the original one if the level size of pyramid is N.
        Then we directly upsample the low-dimension feature maps to get the same size feature as the original feature map via bilinear interpolation.
        Finally, different levels of features are concatenated as the final pyramid pooling global feature.
        Noted that the number of pyramid levels and size of each level can be modified.
        They are related to the size of feature map that is fed into the pyramid pooling layer.
        The structure abstracts different sub-regions by adopting varying-size pooling kernels in a few strides.
        Thus the multi-stage kernels should maintain a reasonable gap in representation.
        Our pyramid pooling module is a four-level one with bin sizes of 1×1, 2×2, 3×3 and 6×6 respectively.
        For the type of pooling operation between max and average, we perform extensive experiments to show the difference in Section 5.2

        The final feature map size is 1/8 of the input image,

        Next, we use the pyramid pooling module (4-levels).
        There are 4-level the pooling kernels
        cover the whole, half of, and small portions of the image. They are
        fused as the global prior. Then we concatenate the prior with the
        original feature map in the final part of (c). It is followed by a
        convolution layer to generate the final prediction map in (d).

        To explain our structure, PSPNet provides an effective global
        contextual prior for pixel-level scene parsing. The pyramid pooling
        module can collect levels of information, more representative than
        global pooling [24]. In terms of computational cost, our PSPNet does
        not much increase it compared to the original dilated FCN network. In
        end-toend learning, the global pyramid pooling module and the local FCN
        feature can be optimized simultaneously

    Ignore:
        >>> from netharn.models.psp import *  # NOQA
        >>> import torch
        >>> self = PSPNet_Resnet50_8s(classes=2).eval()
        >>> inputs = torch.rand(1, 3, 64, 64)
        >>> outputs = self(inputs)
        >>> outputs['class_energy'].shape
    """

    def __init__(self, classes=1000, in_channels=3):
        super(PSPNet_Resnet50_8s, self).__init__()

        import ndsampler
        self.classes = ndsampler.CategoryTree.coerce(classes)

        num_classes = len(self.classes)

        # Load the pretrained weights, remove avg pool
        # layer and get the output stride of 16
        resnet50_8s = resnet50(fully_conv=True, pretrained=True,
                               output_stride=8, remove_avg_pool_layer=True)

        self.psp_head = PSP_head(resnet50_8s.inplanes)

        # Randomly initialize the 1x1 Conv scoring layer
        resnet50_8s.fc = nn.Conv2d(resnet50_8s.inplanes // 4, num_classes, 1)

        self.resnet50_8s = resnet50_8s

    def forward(self, inputs):
        input_spatial_dim = inputs.size()[2:]

        x = inputs
        x = self.resnet50_8s.conv1(x)
        x = self.resnet50_8s.bn1(x)
        x = self.resnet50_8s.relu(x)
        x = self.resnet50_8s.maxpool(x)

        x = self.resnet50_8s.layer1(x)
        x = self.resnet50_8s.layer2(x)
        x = self.resnet50_8s.layer3(x)
        resnet_out = self.resnet50_8s.layer4(x)

        psp_out = self.psp_head(resnet_out)
        fc_out = self.resnet50_8s.fc(psp_out)

        class_energy = upsample(input=fc_out, size=input_spatial_dim)

        outputs = {
            'class_energy': class_energy,
        }
        return outputs

    def _analytic_forward(self, inputs, _OutputFor, _Output, _Hidden,
                          **kwargs):
        """
        Ignore:
            >>> from netharn.models.psp import *  # NOQA
            >>> import netharn as nh
            >>> self = PSPNet_Resnet50_8s(classes=211).eval()
            >>> kwargs = self._analytic_shape_kw()
            >>> globals().update(kwargs)
            >>> inputs = input_shape = (1, 3, 256, 256)
            >>> outputs = self._analytic_forward(inputs, **kwargs)
            >>> print('outputs = {}'.format(ub.repr2(outputs.hidden.shallow(1), nl=-1)))
        """
        try:
            input_spatial_dim = inputs.size()[2:]
        except Exception:
            input_spatial_dim = inputs[2:]

        hidden = _Hidden()

        x = inputs
        hidden['inputs'] = inputs
        x = hidden['conv1'] = _OutputFor(self.resnet50_8s.conv1)(x)
        x = _OutputFor(self.resnet50_8s.bn1)(x)
        x = _OutputFor(self.resnet50_8s.relu)(x)
        x = hidden['maxpool'] = _OutputFor(self.resnet50_8s.maxpool)(x)

        x = hidden['layer1'] = _OutputFor(self.resnet50_8s.layer1)(x)
        x = hidden['layer2'] = _OutputFor(self.resnet50_8s.layer2)(x)
        x = hidden['layer3'] = _OutputFor(self.resnet50_8s.layer3)(x)
        resnet_out = hidden['resnet_out'] = _OutputFor(self.resnet50_8s.layer4)(x)

        psp_out = hidden['psp_out'] = _OutputFor(self.psp_head)(resnet_out)
        fc_out = hidden['fc_out'] = _OutputFor(self.resnet50_8s.fc)(psp_out)

        class_energy = hidden['energy'] = _OutputFor(nn.functional.interpolate)(
            fc_out,
            size=input_spatial_dim,
            mode='bilinear',
            align_corners=True)

        outputs = {
            'class_energy': class_energy,
        }
        outputs = _Output.coerce(outputs, hidden)
        return outputs


def upsample(input, size):
    return nn.functional.interpolate(input=input, size=size, mode='bilinear', align_corners=True)



import six
import torch
import ubelt as ub
from os.path import dirname
from os.path import exists
from os.path import join
from os.path import normpath
from netharn import api
from netharn.initializers.functional import load_partial_state


class Pretrained(api.Initializer, ub.NiceRepr):
    """
    This class initializes a model with pretrained weights from a file on disk.

    If the model topology is slightly different (e.g. the shape of the final
    layer changes), the weights are partially applied. See
    `netharn.initializers.functional.load_partial_state` for mor details on
    this process.

    Attributes:
        fpath (str | PathLike): location of the pretrained weights file.

            This can be a pytorch '.pt' file containing the model state, a path
            to a netharn deploy '.zip' file.

            While it is best practice to use an explicit filepath, we do allow
            `fpath` be a "fuzzy" glob string as long as the pattern resolves to
            a single file, otherwise an error will be thrown.

        leftover (netharn.Initializer | str): backup initializer if the weights
            can only be partially applied. I.E. The initializer applied to the
            leftover weights that were not in the pretrained file. Can either
            be an initializer class or a coercable initializer string.

        mangle (bool, default=True): If True, mangles tensors that have the
            same key, but different shapes forcing them to fit. This might
            destroy information when forcing a a larger tensor into a smaller
            tensor, or leave extra uninitialized room when a small tensor is
            placed in a larger one. Note be careful when mangling a
            classification layer if class indexes are not aligned.

        info (dict, optional): specify explicit history info

        initializer (netharn.Initializer): DEPRECATED use the `leftover`.

    Example:
        >>> from netharn.initializers.pretrained import *
        >>> from netharn.models import toynet
        >>> from os.path import join
        >>> # Save "pretrained" weights to disk
        >>> model1 = toynet.ToyNet2d()
        >>> dpath = ub.ensure_app_cache_dir('netharn', 'tests')
        >>> fpath = join(dpath, 'toynet_weights.pt')
        >>> torch.save(model1.state_dict(), fpath)
        >>> # Create the initializer and point to the pretrained weights
        >>> self = Pretrained(fpath)
        >>> # Apply the pretrained weights to a new model
        >>> model2 = toynet.ToyNet2d()
        >>> self(model2)
        >>> # xdoctest: +SKIP
        >>> # use experimental ubelt features to hash the model state
        >>> ub.util_hash._HASHABLE_EXTENSIONS._register_torch_extensions()
        >>> ub.util_hash._HASHABLE_EXTENSIONS._register_agressive_extensions()
        >>> hash1 = ub.hash_data(model2.state_dict())
        >>> hash2 = ub.hash_data(model1.state_dict())
        >>> assert hash1 == hash2

    Example:
        >>> from netharn.initializers.pretrained import *
        >>> from netharn.models import toynet
        >>> from os.path import join
        >>> # Save "pretrained" weights to disk
        >>> model1 = toynet.ToyNet2d(num_classes=2)
        >>> model2 = toynet.ToyNet2d(num_classes=3)
        >>> dpath = ub.ensure_app_cache_dir('netharn', 'tests')
        >>> fpath = join(dpath, 'toynet_weights1.pt')
        >>> torch.save(model1.state_dict(), fpath)
        >>> # Create the initializer and point to the pretrained weights
        >>> self = Pretrained(fpath, leftover='kaiming_normal')
        >>> # Apply the partial pretrained weights to a new model
        >>> self(model2)
    """
    def __init__(self, fpath, leftover=None, mangle=True, info=None,
                 initializer=None):
        if initializer is not None:
            import warnings
            warnings.warn('Pretrained `initializer` kwarg is deprecated '
                          'in favor of `leftover`', DeprecationWarning)
            leftover = initializer

        self.fpath = fpath
        if isinstance(leftover, six.string_types):
            initializer_ = api.Initializer.coerce(initializer=leftover)
            leftover = initializer_[0](**initializer_[1])

        self.leftover = leftover
        self.mangle = mangle
        self.info = info

    def __nice__(self):
        return self.fpath

    def _rectify_deploy_zip_weights_path(self):
        # Find the path to the weights inside the zipfile
        import zipfile
        fpath = None
        candidates = []
        with zipfile.ZipFile(self.fpath, 'r') as myzip:
            for zinfo in myzip.filelist:
                if zinfo.filename.endswith('deploy_snapshot.pt'):
                    candidates = [zinfo.filename]
                    break
                elif zinfo.filename.endswith('.pt'):
                    candidates.append(zinfo.filename)
        if len(candidates) == 0:
            raise OSError('Cannot find pretrained weights in {}'.format(
                self.fpath))
        elif len(candidates) > 1:
            raise OSError('Multiple weights files in {}'.format(
                self.fpath))
        else:
            fpath = join(self.fpath, candidates[0])
        return fpath

    def _rectify_fpath(self):
        """
        Resolves the `self.fpath`, which may be non-physical path (e.g.
        globstring or zipfile) to an existing physical path if possible.
        """
        if self.fpath is None:
            raise ValueError('Pretrained fpath is None!')
        # Handle torch deployment zipfiles
        if exists(self.fpath) and self.fpath.endswith('.zip'):
            fpath = self._rectify_deploy_zip_weights_path()
        else:
            fpath = self.fpath
            if not exists(fpath) and '*' in fpath:
                import glob
                cands = list(glob.glob(fpath))
                if len(cands) == 1:
                    fpath = cands[0]
                else:
                    raise Exception(
                        'Pattern fpath={!r} must resolve to exactly one file, '
                        'but got cands{!r}'.format(fpath, cands))
        return fpath

    def _load_model_state(self, xpu=None):
        """
        Load the model state from a path or from within a zipfile
        """
        import netharn as nh
        from netharn import XPU

        fpath = self._rectify_fpath()

        if xpu is None:
            xpu = XPU.coerce('cpu')

        try:
            file = nh.util.zopen(fpath, 'rb', seekable=True)
            model_state_dict = xpu.load(file)
        except Exception:
            print('Failed to open fpath = {!r}'.format(fpath))
            raise
        return model_state_dict

    def forward(self, model, verbose=2):
        """
        Apply the pretrained weights to the model
        """
        from netharn import XPU
        xpu = XPU.from_data(model)

        model_state_dict = self._load_model_state(xpu=xpu)

        if 'model_state_dict' in model_state_dict:
            model_state_dict = model_state_dict['model_state_dict']
        elif 'state_dict' in model_state_dict:
            model_state_dict = model_state_dict['state_dict']
        elif 'weights' in model_state_dict:
            model_state_dict = model_state_dict['weights']
        else:
            # If the dictionary is flat (i.e. all values are tensors) then it
            # is safe to assume this file only contains weights.
            # Otherwise raise an exception.
            if not all(torch.is_tensor(v) for v in model_state_dict.values()):
                raise Exception(
                    'snapshot file is nested, but does not have expected keys: '
                    'model_state_dict or weights. Root keys are {}'.format(
                        sorted(model_state_dict.keys())
                    ))
        # Remove any DataParallel / DataSerial
        raw_model = xpu.raw(model)
        info = load_partial_state(raw_model, model_state_dict,
                                  leftover=self.leftover,
                                  mangle=self.mangle,
                                  verbose=verbose)
        return info

    def history(self):
        """
        if available return the history of the model as well
        """
        import netharn as nh
        if self.info is None:
            # TODO: check for train_info.json in a few different places
            fpath = self._rectify_fpath()
            snap_fpath = ub.expandpath(fpath)
            candidate_paths = [
                join(dirname(snap_fpath), 'train_info.json'),
                join(dirname(dirname(snap_fpath)), 'train_info.json'),
            ]
            info = None
            for info_fpath in candidate_paths:
                info_fpath = normpath(info_fpath)
                try:
                    # Info might be inside of a zipfile
                    info = nh.util.read_json(nh.util.zopen(info_fpath))
                    break
                except Exception:
                    pass
            if info is None:
                info = '__UNKNOWN__'
        else:
            info = self.info
        return info
